/*
  game_alloc.c
  jigplay-macos
 
  Created by François Pessaux on 16/09/10.
  Copyright 2010. All rights reserved.
*/

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#ifdef __JIGPLAY_IOS__
#include <CoreGraphics/CGContext.h>
#include <CoreGraphics/CGBitmapContext.h>
#endif

#include "misc.h"
#include "piece.h"
#include "game_struct.h"
#include "game_alloc.h"
#include "game_free.h"
#include "game_manips.h"
#include "texturemap.h"
#include "alerts_n_errors.h"



/** Insert [edge] after the edge [after] in the double linked list of edges
 sorted by increasing minimal x coordinate. This function is used
 at list initialization-time. It doesn't check the coordinates, so the
 fact that edges as they are inserted are well sorted is uniquely ensured
 by the algorithm that calls this function. */
static void insertAfterInXList (struct edge *edge, struct edge *after)
{
  edge->x_sorted_next = after->x_sorted_next ;
  after->x_sorted_next = edge ;
  if (edge->x_sorted_next != NULL)
    edge->x_sorted_next->x_sorted_prev = edge ;
  edge->x_sorted_prev = after ;
}



/** Insert [edge] after the edge [after] in the double linked list of edges
 sorted by increasing minimal y coordinate. This function is used
 at list initialization-time. It doesn't check the coordinates, so the
 fact that edges as they are inserted are well sorted is uniquely ensured
 by the algorithm that calls this function. */
static void InsertAfterInYList (struct edge *edge, struct edge *after)
{
  edge->y_sorted_next = after->y_sorted_next ;
  after->y_sorted_next = edge ;
  if (edge->y_sorted_next != NULL)
    edge->y_sorted_next->y_sorted_prev = edge ;
  edge->y_sorted_prev = after ;
}



/** Highly inspired from NetSketch code snippet published on
    http://www.gotow.net/creative/wordpress/?p=7
    This function makes a copy of the image passed as argument, scaling it
    according to the specified factor. By the way we are making a copy, we
    make so that the result image has an alpha channel. Since we anyway need
    to create a new bitmap context with CGBitmapContextCreate, it is not more
    expensive to enforce transparency, this way we don't need to check later
    if the original image had transparency and if not add it calling
    CGBitmapContextCreate again in the function copyImageAndAddAlphaChannel. */
CGImageRef makeScaledCopyOfImageAndAddAlphaChannel (CGImageRef image,
                                                    float scale_factor)
{
  /* Result image. */
  CGImageRef scaled_image ;
  CGContextRef offscreen_context ;
  /* Image buffer where we will write the result of the scaling. */
  unsigned char *bitmap_data ;
  /* Number of bytes in the image buffer. */
  int bitmap_size ;
  int nb_bytes_per_row ;
  int scaled_width = (int) ((float) CGImageGetWidth (image) * scale_factor) ;
  int scaled_height = (int) ((float) CGImageGetHeight (image) * scale_factor) ;
  
  /* Since we want 32 bits (8 bits for each of red, green, blue, and alpha)
     the number of bytes per row is 4 times the scaled width of image. */
  nb_bytes_per_row = scaled_width * 4 ;
  bitmap_size = nb_bytes_per_row * scaled_height ;
  
  bitmap_data = malloc (bitmap_size * sizeof (unsigned char)) ;
  if (bitmap_data == NULL) return (NULL) ;
  
  CGColorSpaceRef colorspace = CGColorSpaceCreateDeviceRGB () ;
  
  /* Create the bitmap context. We want pre-multiplied ARGB, 8-bits per
     component. Regardless of what the source image format is (CMYK, Grayscale,
     and so on) it will be converted over to the format specified here by
     CGBitmapContextCreate. */
  offscreen_context =
    CGBitmapContextCreate
      (bitmap_data, scaled_width, scaled_height, 8, nb_bytes_per_row,
       colorspace, kCGImageAlphaPremultipliedFirst) ;
  CGColorSpaceRelease (colorspace) ;
  
  if (offscreen_context == NULL) {
    free (bitmap_data) ;
    return (NULL) ;
  }
  
  /* Draw the image to the bitmap context. Once we draw, the memory allocated
     for the context for rendering will then contain the raw image data in the
     specified color space. */
  CGContextDrawImage
    (offscreen_context, CGRectMake (0, 0, scaled_width, scaled_height), image) ;
  
  scaled_image = CGBitmapContextCreateImage (offscreen_context) ;
  CGContextRelease (offscreen_context) ;
  free (bitmap_data) ;
  
  return (scaled_image) ;
}



/** Highly inspired from the code snippet published on
    http://copytaste.com/pa18c33z */
CGImageRef copyImageAndAddAlphaChannel (CGImageRef sourceImage)
{ 
  CGImageRef copied_image = NULL ;
  int width, height ;
  CGColorSpaceRef color_space ;
  CGContextRef offscreen_context ;
  
  width = CGImageGetWidth (sourceImage) ; 
  height = CGImageGetHeight (sourceImage) ; 
  color_space = CGColorSpaceCreateDeviceRGB () ;
  
  offscreen_context =
    CGBitmapContextCreate
      (NULL, width, height, 8, 0, color_space, kCGImageAlphaPremultipliedFirst) ;
  
  if (offscreen_context != NULL) { 
    CGContextDrawImage
      (offscreen_context, CGRectMake (0, 0, width, height), sourceImage) ;
    
    copied_image = CGBitmapContextCreateImage (offscreen_context) ;
    CGContextRelease (offscreen_context) ;
  } 
  
  CGColorSpaceRelease (color_space) ;
  return (copied_image) ;
}



#ifdef __JIGPLAY_MAC_OS__
/** Loads an image from a file name and returns by side effect its width and
    height. Allowed formats are JPEG and PNG. */
static CGImageRef loadGameImage (const char *image_file_name,
                                 float scale_factor, int *width, int *height)
#else
#ifdef __JIGPLAY_IOS__
static CGImageRef loadGameImage (CGImageRef from_image,
                                 float scale_factor, int *width, int *height)
#endif
#endif
{
  CGImageRef image = NULL ;
  CGImageAlphaInfo alphaInfo ;
#ifdef __JIGPLAY_MAC_OS__
  char *suffix ;
  CGDataProviderRef provider = NULL ;
  int formatIsJpeg ;
  
  /* Check the format of the image file from its extension. */
  suffix = rindex (image_file_name, (int) '.') ;
  if (suffix == NULL) return (NULL) ;
  if ((strcasecmp (".jpg", suffix) == 0) ||
      (strcasecmp (".jpeg", suffix) == 0)) {
    /* A priori, format is JPEG. */
    formatIsJpeg = TRUE ;
  }
  else {
    if (strcasecmp (".png", suffix) == 0) {
      /* A priori, format is PNG. */
      formatIsJpeg = FALSE ;
    }
    else {
      /* A priori, non-supported format. */
      return (NULL) ;
    }
  }
  
  provider = CGDataProviderCreateWithFilename (image_file_name) ;
  if (provider == NULL) {
    if (width != NULL) *width = 0 ;
    if (height != NULL) *height = 0 ;
    return (NULL) ;
  }

  if (formatIsJpeg) {
    image =
      CGImageCreateWithJPEGDataProvider
        (provider, NULL, true, kCGRenderingIntentDefault) ;
  }
  else {
    image =
      CGImageCreateWithPNGDataProvider
        (provider, NULL, true, kCGRenderingIntentDefault) ;
  }

  CGDataProviderRelease (provider) ;

  if (image == NULL) {
    if (width != NULL) *width = 0 ;
    if (height != NULL) *height = 0 ;
    return (NULL) ;
  }
#else
#ifdef __JIGPLAY_IOS__
  image = from_image ;
#endif
#endif
  
  if (width != NULL) *width = CGImageGetWidth (image) ;
  if (height != NULL) *height = CGImageGetHeight (image) ;
  
  /* Modify the image size according to the scaling factor if needed and
     either scale the image and add transparency or simply check if we need to
     add transparency. */
  if (scale_factor > 0) {
    CGImageRef tmp = image ;
    *width *= scale_factor ;
    *height *= scale_factor ;
    /* No need to check if we need to add transparency, the scaling routine
       always add it by the way because this doesn't cost it more. */
    image = makeScaledCopyOfImageAndAddAlphaChannel (tmp, scale_factor) ;
    CGImageRelease (tmp) ;
  }
  else {
    /* If the image doesn't have any transparency channel, add one. */
    alphaInfo = CGImageGetAlphaInfo (image) ;
    if ((alphaInfo == kCGImageAlphaNone) ||
        (alphaInfo == kCGImageAlphaNoneSkipLast) ||
        (alphaInfo == kCGImageAlphaNoneSkipFirst)) {
      CGImageRef tmp = image ;
      image = copyImageAndAddAlphaChannel (tmp) ;
      CGImageRelease (tmp) ;
    }
  }

  return (image) ;
}



unsigned char* rawLoadImageFile (const char *image_file_name, int *size)
{
  int file_size ;
  unsigned char *data ;
  FILE *in ;
  
  in = fopen (image_file_name, "rb") ;
  if (in == NULL) return (NULL) ;
  
  /* Guess the size of the file by setting the position indicator at the end of
     the file and asking its offset. Then rewind to the beginning of the
     file. */
  fseek (in, 0, SEEK_END) ;
  file_size = ftell (in) ;
  fseek (in, 0, SEEK_SET) ;
  
  data = malloc (file_size * sizeof (unsigned char)) ;
  if (data == NULL) {
    fclose (in) ;
    *size = 0 ;
    return (NULL) ;
  }
  
  fread (data, sizeof (unsigned char), file_size, in) ;
  
  fclose (in) ;
  *size = file_size ;
  return (data) ;
}



/** Initiates a new game creation.
    Note that if scale_factor is negative or null, then this means that no
    scaling must be applied to the image. */
#ifdef __JIGPLAY_MAC_OS__
struct game* newGame (const char *image_filename, int screen_width,
                      int screen_height, int nb_columns_of_pieces,
                      int nb_lines_of_pieces, float scale_factor)
#else
#ifdef __JIGPLAY_IOS__
struct game* newGame (CGImageRef from_image, int screen_width,
                      int screen_height, int nb_columns_of_pieces,
                      int nb_lines_of_pieces, float scale_factor)
#endif
#endif
{
  int x, y, i ;
  int piece_width, piece_height ;
  int layout_nb_columns, layout_nb_lines ;
  int image_width, image_height ;
  /* Matrix representing the pieces of the jigsaw in term of column/line
     position. We allocate this matrix as a contiguous array and index it by
     (y * nb_columns) + x. This array contains the pieces whose
     address will be shared everywhere. */
  struct piece *pieces = NULL ;
  /* Matrix containing pieces that are (even partially) in the cell
     corresponding to the indices of this cell. The layout has many
     rows and columns the jigsaw has currently but could be as large
     as the playfield is. */
  struct layout_cell *layout = NULL ;
  /* Temporary edge pointer for various usages. */
  struct edge *tmp_edge ;
  /* Pointer used to initialize the sorted double linked lists. */
  struct edge *prev ;
  CGImageRef game_image ;
  int raw_image_file_size ;
  unsigned char *raw_image_file_data = NULL ;
  /* The final game structure we will return. */
  struct game *game = NULL ;

#ifdef __JIGPLAY_MAC_OS__
  game_image =
    loadGameImage (image_filename, scale_factor, &image_width, &image_height) ;
  if (game_image == NULL) {
    displayMessage ("Error", "Unable to load image file.", AC_Critical) ;
    return (NULL) ;
  }
  
  raw_image_file_data =
    rawLoadImageFile (image_filename, &raw_image_file_size) ;
  if (raw_image_file_data == NULL) {
    CGImageRelease (game_image) ;
    displayMessage ("Error", "Unable to load image file.", AC_Critical) ;
    return (NULL) ;
  }  
#else
#ifdef __JIGPLAY_IOS__
#endif
  game_image =
    loadGameImage (from_image, scale_factor, &image_width, &image_height) ;
  if (game_image == NULL) {
    displayMessage ("Error", "Unable to load image file.", AC_Critical) ;
    return (NULL) ;
  }
  
  // TODO image data
#endif
  
  /* Compute the pieces width and height according to the number of
     columns and lines of pieces requested and the size of the
     image. */
  piece_width = image_width / nb_columns_of_pieces ;
  piece_height = image_height / nb_lines_of_pieces ;
  
  /* Attention, if the image size is not exactly divisible by the number of
     pieces, we must crop the image size to make so that the number of pieces
     times the pieces size is exactly the image size. */
  image_width = nb_columns_of_pieces * piece_width ;
  image_height = nb_lines_of_pieces * piece_height ;
  
  /* Allocate all the pieces of the jigsaw, initially seen as a matrix
     of size [nb_columns_of_pieces * nb_lines_of_pieces]. */
  pieces =
    malloc
      ((nb_columns_of_pieces * nb_lines_of_pieces) * sizeof (struct piece)) ;
  if (pieces == NULL) {
    CGImageRelease (game_image) ;
    free (raw_image_file_data) ;
    displayMessage ("Error", "Not enough memory.", AC_Critical) ;
    return (NULL) ;
  }
  memset
    (pieces, 0,
     (nb_columns_of_pieces * nb_lines_of_pieces) * sizeof (struct piece)) ;
  
  /* Now, allocate the layout. Note that the layout must contain at least
     as many cells than the matrix of pieces has otherwise the pieces of
     the jigsaw would't fit inside.
     Moreover, to let the player move pieces on an area wider than the
     jigsaw's image, we make the layout bigger and use all the window's
     surface. */
  layout_nb_columns = screen_width / piece_width ;
  if ((screen_width % piece_width) != 0) layout_nb_columns++ ;
  layout_nb_columns = max (layout_nb_columns, nb_columns_of_pieces) ;
  layout_nb_lines = screen_height / piece_height ;
  if ((screen_height % piece_height) != 0) layout_nb_lines++ ;
  layout_nb_lines = max (layout_nb_lines, nb_lines_of_pieces) ;
  layout =
    malloc
      ((layout_nb_columns * layout_nb_lines) * sizeof (struct layout_cell)) ;
  if (layout == NULL) {
    CGImageRelease (game_image) ;
    free (raw_image_file_data) ;
    free (pieces) ;
    displayMessage ("Error", "Not enough memory.", AC_Critical) ;
    return (NULL) ;
  }
  
  /* Initialize the layout. */
  for (i = 0; i < (layout_nb_columns * layout_nb_lines); i++) {
    layout[i].to_redraw = 0 ;
    layout[i].hosted_pieces = NULL ;
  }

  /* Loop for basic creation of the pieces, some parts will be
     initialized to default value and update later when more
     information will be available. */
  for (x = 0; x < nb_columns_of_pieces; x++) {
    for (y = 0; y < nb_lines_of_pieces; y++) {
      int k ;
      int piece_index_and_id = (y * nb_columns_of_pieces) + x ;
      
      /* Keep a pointer to the current piece in the pieces matrix. */
      struct piece *piece = &pieces[piece_index_and_id] ;

      /* Compute the id of the first edge of the piece. */
      int first_edge_debug_id = ((x * nb_lines_of_pieces) + y) * 4 ;

      /* Initialize the piece's structure. Note that id is equal to index
         in the array of pieces. */
      piece->piece_debug_id = piece_index_and_id ;
      piece->center.x =  (x * piece_width) + (piece_width / 2) ;
      piece->center.y = (y * piece_height) + (piece_height / 2) ;
      /* Bounding box will be initialized once the splines are computed. For
         the moment, setting it to (0, 0 - (0, 0) is fine since this corresponds
         to a 0-wide bounding box centered around the piece's center. */
      piece->delta_center_to_top_left.x = 0 ;
      piece->delta_center_to_top_left.y = 0 ;
      piece->delta_center_to_bottom_right.x = 0 ;
      piece->delta_center_to_bottom_right.y = 0 ;
      
      /* Allocate memory for the 4 edges. */
      for (k = 0; k < 4; k++) {
        piece->free_edges[k] = malloc (sizeof (struct edge)) ;
        if (piece->free_edges[k] == NULL) {
          CGImageRelease (game_image) ;
          free (raw_image_file_data) ;
          freeLayout (layout, layout_nb_columns * layout_nb_lines) ;
          freePieces (pieces, nb_columns_of_pieces * nb_lines_of_pieces) ;
          displayMessage ("Error", "Not enough memory.", AC_Critical) ;
          return (NULL) ;
        }
      }    /* End of for (k = 0; k < 4; k++). */
        
      /* To be fully initialized below once edges exist. */
      piece->free_edges[NORTH]->active = FALSE ;
      piece->free_edges[EAST]->active = FALSE ;
      piece->free_edges[SOUTH]->active = FALSE ;
      piece->free_edges[WEST]->active = FALSE ;
      /* Initialize the connected set with a fresh one only containing
         the current piece. */
      {
        struct connected_set *con_set ;
        struct pieces_list *p_lst ;

        con_set = malloc (sizeof (struct connected_set)) ;
        if (con_set == NULL) {
          CGImageRelease (game_image) ;
          free (raw_image_file_data) ;
          freeLayout (layout, layout_nb_columns * layout_nb_lines) ;
          freePieces (pieces, nb_columns_of_pieces * nb_lines_of_pieces) ;
          displayMessage ("Error", "Not enough memory.", AC_Critical) ;
          return (NULL) ;
        }
        
        p_lst = malloc (sizeof (struct pieces_list)) ;
        if (p_lst == NULL) {
          free (con_set) ;
          CGImageRelease (game_image) ;
          free (raw_image_file_data) ;
          freeLayout (layout, layout_nb_columns * layout_nb_lines) ;
          freePieces (pieces, nb_columns_of_pieces * nb_lines_of_pieces) ;
          displayMessage ("Error", "Not enough memory.", AC_Critical) ;
          return (NULL) ;
        }
        
        p_lst->next = NULL ;
        p_lst->piece = piece ;
        con_set->members = p_lst ;
        con_set->refcount = 1 ;
        piece->connected_pieces = con_set ;
      }
      /* Image will be initialized later. */
      piece->piece_image = NULL ;

      /* Now, create the edges of the piece. */

      /* North. */
      piece->free_edges[NORTH]->active = 1 ;
      piece->free_edges[NORTH]->p0.x = x * piece_width ;
      piece->free_edges[NORTH]->p0.y = y * piece_height ;
      piece->free_edges[NORTH]->p0_anchor_id =
        x + y * (nb_columns_of_pieces + 1) ;
      piece->free_edges[NORTH]->p1.x = (x + 1) * piece_width - 1 ;
      piece->free_edges[NORTH]->p1.y = y * piece_height ;
      piece->free_edges[NORTH]->p1_anchor_id =
        1 + x + y * (nb_columns_of_pieces + 1) ;
      piece->free_edges[NORTH]->spline = NULL ; /* To be initialized later. */
      piece->free_edges[NORTH]->part_of_piece = piece ;
      piece->free_edges[NORTH]->x_sorted_prev = NULL ;
      piece->free_edges[NORTH]->x_sorted_next = NULL ;
      piece->free_edges[NORTH]->y_sorted_prev = NULL ;
      piece->free_edges[NORTH]->y_sorted_next = NULL ;
      piece->free_edges[NORTH]->id = first_edge_debug_id ;

      /* East. */
      piece->free_edges[EAST]->active = 1 ;
      piece->free_edges[EAST]->p0.x = (x + 1) * piece_width - 1 ;
      piece->free_edges[EAST]->p0.y = y * piece_height ;
      piece->free_edges[EAST]->p0_anchor_id =
        1 + x + y * (nb_columns_of_pieces + 1) ;
      piece->free_edges[EAST]->p1.x = (x + 1) * piece_width - 1 ;
      piece->free_edges[EAST]->p1.y = (y + 1) * piece_height - 1 ;
      piece->free_edges[EAST]->p1_anchor_id =
        1 + x + (y + 1) * (nb_columns_of_pieces + 1) ;
      piece->free_edges[EAST]->spline = NULL ; /* To be initialized later. */
      piece->free_edges[EAST]->part_of_piece = piece ;
      piece->free_edges[EAST]->x_sorted_prev = NULL ;
      piece->free_edges[EAST]->x_sorted_next = NULL ;
      piece->free_edges[EAST]->y_sorted_prev = NULL ;
      piece->free_edges[EAST]->y_sorted_next = NULL ;
      piece->free_edges[EAST]->id = first_edge_debug_id + 1 ;

      /* South. */
      piece->free_edges[SOUTH]->active = 1 ;
      piece->free_edges[SOUTH]->p0.x = (x + 1) * piece_width - 1 ;
      piece->free_edges[SOUTH]->p0.y = (y + 1) * piece_height - 1 ;
      piece->free_edges[SOUTH]->p0_anchor_id =
        1 + x + (y + 1) * (nb_columns_of_pieces + 1) ;
      piece->free_edges[SOUTH]->p1.x = x * piece_width ;
      piece->free_edges[SOUTH]->p1.y = (y + 1) * piece_height - 1 ;
      piece->free_edges[SOUTH]->p1_anchor_id =
        x + (y + 1) * (nb_columns_of_pieces + 1) ;
      piece->free_edges[SOUTH]->spline = NULL ; /* To be initialized later. */
      piece->free_edges[SOUTH]->part_of_piece = piece ;
      piece->free_edges[SOUTH]->x_sorted_prev = NULL ;
      piece->free_edges[SOUTH]->x_sorted_next = NULL ;
      piece->free_edges[SOUTH]->y_sorted_prev = NULL ;
      piece->free_edges[SOUTH]->y_sorted_next = NULL ;
      piece->free_edges[SOUTH]->id = first_edge_debug_id + 2 ;

      /* West. */
      piece->free_edges[WEST]->active = 1 ;
      piece->free_edges[WEST]->p0.x = x * piece_width ;
      piece->free_edges[WEST]->p0.y = (y + 1) * piece_height - 1 ;
      piece->free_edges[WEST]->p0_anchor_id =
        x + (y + 1) * (nb_columns_of_pieces + 1) ;
      piece->free_edges[WEST]->p1.x = x * piece_width ;
      piece->free_edges[WEST]->p1.y = y * piece_height ;
      piece->free_edges[WEST]->p1_anchor_id =
        x + y * (nb_columns_of_pieces + 1) ;
      piece->free_edges[WEST]->spline = NULL ; /* To be initialized later. */
      piece->free_edges[WEST]->part_of_piece = piece ;
      piece->free_edges[WEST]->x_sorted_prev = NULL ;
      piece->free_edges[WEST]->x_sorted_next = NULL ;
      piece->free_edges[WEST]->y_sorted_prev = NULL ;
      piece->free_edges[WEST]->y_sorted_next = NULL ;
      piece->free_edges[WEST]->id = first_edge_debug_id + 3 ;
    }    /* End of for (y = 0; y < nb_lines; y++). */
  }      /* End of for (x = 0; x < nb_columns; x++). */

  /* Now, splines updates for all pieces. */

  /* Update the splines north and south by splitting by columns. */
  for (x = 0; x < nb_columns_of_pieces; x++) {
    /* Just used to store the index of a piece in the pieces matrix. */
    int index ;
    /* Some temporary variables. */
    struct point *tmp_point_ptr ;
    struct spline *tmp_spline_ptr ;

    /* Init of the north of each cell of the first line. */
    index = x ;       /* pieces[x][0]. */
    tmp_point_ptr = malloc (2 * sizeof (struct point)) ;
    if (tmp_point_ptr == NULL) {
      CGImageRelease (game_image) ;
      free (raw_image_file_data) ;
      freeLayout (layout, layout_nb_columns * layout_nb_lines) ;
      freePieces (pieces, nb_columns_of_pieces * nb_lines_of_pieces) ;
      displayMessage ("Error", "Not enough memory.", AC_Critical) ;
      return (NULL) ;
    }
    
    tmp_spline_ptr = malloc (sizeof (struct spline)) ;
    if (tmp_spline_ptr == NULL) {
      free (tmp_point_ptr) ;
      CGImageRelease (game_image) ;
      free (raw_image_file_data) ;
      freeLayout (layout, layout_nb_columns * layout_nb_lines) ;
      freePieces (pieces, nb_columns_of_pieces * nb_lines_of_pieces) ;
      displayMessage ("Error", "Not enough memory.", AC_Critical) ;
      return (NULL) ;
    }
    
    tmp_spline_ptr->nb_points = 2 ;
    tmp_spline_ptr->vertices = tmp_point_ptr ;
    pieces[index].free_edges[NORTH]->spline = tmp_spline_ptr ;
    tmp_point_ptr[0].x = x * piece_width ;
    tmp_point_ptr[0].y = 0 ;
    tmp_point_ptr[1].x = (x + 1) * piece_width - 1 ;
    tmp_point_ptr[1].y = 0 ;
    pieces[index].delta_center_to_top_left.x = - (piece_width / 2) ;
    pieces[index].delta_center_to_top_left.y = - (piece_height / 2) ;
    pieces[index].delta_center_to_bottom_right.x = piece_width / 2 ;
    pieces[index].delta_center_to_bottom_right.y = - (piece_height / 2) ;
    
    /* Init of the south of each cell of the last line. */
    /* pieces[x][nb_lines - 1]. */
    index = ((nb_lines_of_pieces - 1) * nb_columns_of_pieces) + x ;
    tmp_point_ptr = malloc (2 * sizeof (struct point)) ;
    if (tmp_point_ptr == NULL) {
      CGImageRelease (game_image) ;
      free (raw_image_file_data) ;
      freeLayout (layout, layout_nb_columns * layout_nb_lines) ;
      freePieces (pieces, nb_columns_of_pieces * nb_lines_of_pieces) ;
      displayMessage ("Error", "Not enough memory.", AC_Critical) ;
      return (NULL) ;
    }
    
    tmp_spline_ptr = malloc (sizeof (struct spline)) ;
    if (tmp_spline_ptr == NULL) {
      free( tmp_point_ptr) ;
      CGImageRelease (game_image) ;
      free (raw_image_file_data) ;
      freeLayout (layout, layout_nb_columns * layout_nb_lines) ;
      freePieces (pieces, nb_columns_of_pieces * nb_lines_of_pieces) ;
      displayMessage ("Error", "Not enough memory.", AC_Critical) ;
      return (NULL) ;
    }
    
    tmp_spline_ptr->nb_points = 2 ;
    tmp_spline_ptr->vertices = tmp_point_ptr ;
    pieces[index].free_edges[SOUTH]->spline = tmp_spline_ptr ;
    tmp_point_ptr[0].x = (x + 1) * piece_width - 1 ;
    tmp_point_ptr[0].y = image_height - 1 ;
    tmp_point_ptr[1].x = x * piece_width ;
    tmp_point_ptr[1].y = image_height - 1 ;
    pieces[index].delta_center_to_top_left.x = - (piece_width / 2) ;
    pieces[index].delta_center_to_top_left.y = piece_height / 2 ;
    pieces[index].delta_center_to_bottom_right.x = piece_width / 2 ;
    pieces[index].delta_center_to_bottom_right.y = piece_height / 2 ;
    
    /* Create the horizontal splines of the column. */
    for (y = 1; y < nb_lines_of_pieces; y++) {
      /* Points used as out-mode parameter to receive the spline's
         bounding-box when a spline is created. */
      struct point spline_top_left, spline_bottom_right ;
      struct point *horiz_segments ;
      struct spline *horiz_spline ;

      /* Horizontal spline, hence north of the current cell. */
      horiz_segments =
        makeHorizontalEdgeSegments
          ((x * piece_width), (y * piece_height), (piece_width - 1)) ;
      if (horiz_segments == NULL) {
        CGImageRelease (game_image) ;
        free (raw_image_file_data) ;
        freeLayout (layout, layout_nb_columns * layout_nb_lines) ;
        freePieces (pieces, nb_columns_of_pieces * nb_lines_of_pieces) ;        
        displayMessage ("Error", "Not enough memory.", AC_Critical) ;
        return (NULL) ;
      }
      
      horiz_spline =
        buildSpline
          (1, horiz_segments, NB_UNIT_EDGE_POINTS, &spline_top_left,
           &spline_bottom_right) ;
      if (horiz_spline == NULL) {
        free (horiz_segments) ;
        CGImageRelease (game_image) ;
        free (raw_image_file_data) ;
        freeLayout (layout, layout_nb_columns * layout_nb_lines) ;
        freePieces (pieces, nb_columns_of_pieces * nb_lines_of_pieces) ;        
        displayMessage ("Error", "Not enough memory.", AC_Critical) ;
        return (NULL) ;
      }
      
      /* Release the segments, they are useless now. */
      free (horiz_segments) ;
      
      /* Init the north of the current cell. */
      index = (y * nb_columns_of_pieces) + x ;    /* pieces [x][y]. */
      pieces[index].free_edges[NORTH]->spline = horiz_spline ;
      pieces[index].delta_center_to_top_left.x =
        min
         (pieces[index].delta_center_to_top_left.x,
          (spline_top_left.x - pieces[index].center.x)) ;
      pieces[index].delta_center_to_top_left.y =
        min
          (pieces[index].delta_center_to_top_left.y,
           (spline_top_left.y - pieces[index].center.y)) ;
      pieces[index].delta_center_to_bottom_right.x =
        max
          (pieces[index].delta_center_to_bottom_right.x,
           (spline_bottom_right.x - pieces[index].center.x)) ;
      pieces[index].delta_center_to_bottom_right.y =
        max
        (pieces[index].delta_center_to_bottom_right.y,
         (spline_bottom_right.y - pieces[index].center.y)) ;
      
      /* Init of the south of the y previous cell. */
      if (y > 0) {
        /* pieces[x][y - 1]. */
        index = ((y - 1) * nb_columns_of_pieces) + x ;
        pieces[index].free_edges[SOUTH]->spline =
          build_opposite_spline (/* shift_up: */ 1, horiz_spline) ;
        pieces[index].delta_center_to_top_left.x =
          min
            (pieces[index].delta_center_to_top_left.x,
             (spline_top_left.x - pieces[index].center.x)) ;
        pieces[index].delta_center_to_top_left.y =
          min
            (pieces[index].delta_center_to_top_left.y,
             (spline_top_left.y - pieces[index].center.y)) ;
        pieces[index].delta_center_to_bottom_right.x =
          max
            (pieces[index].delta_center_to_bottom_right.x,
             (spline_bottom_right.x - pieces[index].center.x)) ;
        pieces[index].delta_center_to_bottom_right.y =
          max
            (pieces[index].delta_center_to_bottom_right.y,
             (spline_bottom_right.y - pieces[index].center.y)) ;
      }   /* End of if (y > 0). */
    }     /* End of for (y = 1; y < nb_lines; y++). */
  }       /* End of for (x = 0; x < to nb_columns; x++). */

  /* Update the splines east and west by splitting by lines. */
  for (y = 0; y < nb_lines_of_pieces; y++) {
    /* Just used to store the index of a piece in the pieces matrix. */
    int index ;
    /* Some temporary variables. */
    struct point *tmp_point_ptr ;
    struct spline *tmp_spline_ptr ;

    /* Init of the west of each cell of the first column. */
    /* pieces[0][y]. */
    index = y * nb_columns_of_pieces ;
    tmp_point_ptr = malloc (2 * sizeof (struct point)) ;
    if (tmp_point_ptr == NULL) {
      CGImageRelease (game_image) ;
      free (raw_image_file_data) ;
      freeLayout (layout, layout_nb_columns * layout_nb_lines) ;
      freePieces (pieces, nb_columns_of_pieces * nb_lines_of_pieces) ;
      displayMessage ("Error", "Not enough memory.", AC_Critical) ;
      return (NULL) ;
    }
    
    tmp_spline_ptr = malloc (sizeof (struct spline)) ;
    if (tmp_spline_ptr == NULL) {
      free (tmp_point_ptr) ;
      CGImageRelease (game_image) ;
      free (raw_image_file_data) ;
      freeLayout (layout, layout_nb_columns * layout_nb_lines) ;
      freePieces (pieces, nb_columns_of_pieces * nb_lines_of_pieces) ;
      displayMessage ("Error", "Not enough memory.", AC_Critical) ;
      return (NULL) ;
    }
    
    tmp_spline_ptr->nb_points = 2 ;
    tmp_spline_ptr->vertices = tmp_point_ptr ;
    pieces[index].free_edges[WEST]->spline = tmp_spline_ptr ;
    tmp_point_ptr[0].x = 0 ;
    tmp_point_ptr[0].y = (y + 1) * piece_height - 1 ;
    tmp_point_ptr[1].x = 0 ;
    tmp_point_ptr[1].y = y * piece_height ;
    pieces[index].delta_center_to_top_left.x =
      min
        (pieces[index].delta_center_to_top_left.x,
         (- (piece_width / 2))) ;
    pieces[index].delta_center_to_top_left.y =
      min
        (pieces[index].delta_center_to_top_left.y,
         (- (piece_height / 2))) ;
    pieces[index].delta_center_to_bottom_right.x =
      max
        (pieces[index].delta_center_to_bottom_right.x,
         (- (piece_width / 2))) ;
    pieces[index].delta_center_to_bottom_right.y =
      max
        (pieces[index].delta_center_to_bottom_right.y,
         (piece_height / 2)) ;
    
    /* Init of the east of each cell of the last column. */
    /* pieces[nb_columns - 1][y]. */
    index = (y * nb_columns_of_pieces) + (nb_columns_of_pieces - 1) ;
    tmp_point_ptr = malloc (2 * sizeof (struct point)) ;
    if (tmp_point_ptr == NULL) {
      CGImageRelease (game_image) ;
      free (raw_image_file_data) ;
      freeLayout (layout, layout_nb_columns * layout_nb_lines) ;
      freePieces (pieces, nb_columns_of_pieces * nb_lines_of_pieces) ;
      displayMessage ("Error", "Not enough memory.", AC_Critical) ;
      return (NULL) ;
    }
    tmp_spline_ptr = malloc (sizeof (struct spline)) ;
    if (tmp_spline_ptr == NULL) {
      free (tmp_point_ptr) ;
      CGImageRelease (game_image) ;
      free (raw_image_file_data) ;
      freeLayout (layout, layout_nb_columns * layout_nb_lines) ;
      freePieces (pieces, nb_columns_of_pieces * nb_lines_of_pieces) ;
      displayMessage ("Error", "Not enough memory.", AC_Critical) ;
      return (NULL) ;
    }
    tmp_spline_ptr->nb_points = 2 ;
    tmp_spline_ptr->vertices = tmp_point_ptr ;
    pieces[index].free_edges[EAST]->spline = tmp_spline_ptr ;
    tmp_point_ptr[0].x = image_width - 1 ;
    tmp_point_ptr[0].y = y * piece_height ;
    tmp_point_ptr[1].x = image_width - 1 ;
    tmp_point_ptr[1].y = (y + 1) * piece_height - 1 ;
    pieces[index].delta_center_to_top_left.x =
      min
        (pieces[index].delta_center_to_top_left.x,
         (piece_width / 2)) ;
    pieces[index].delta_center_to_top_left.y =
      min
        (pieces[index].delta_center_to_top_left.y,
         (- (piece_height / 2))) ;
    pieces[index].delta_center_to_bottom_right.x =
      max
        (pieces[index].delta_center_to_bottom_right.x,
         (piece_width / 2)) ;
    pieces[index].delta_center_to_bottom_right.y =
      max
        (pieces[index].delta_center_to_bottom_right.y,
         (piece_height / 2)) ;
    
    /* Create the vertical splines of the line. */
    for (x = 1; x < nb_columns_of_pieces; x++) {
      /* Points used as out-mode parameter to receive the spline's
         bounding-box when a spline is created. */
      struct point spline_top_left, spline_bottom_right ;
      struct point *vert_segments ;
      struct spline *vert_spline ;

      /* Vertical spline, hence west of the current cell. */
      vert_segments =
        makeVerticalEdgeSegments
        ((x * piece_width), (y * piece_height), (piece_height - 1)) ;
      if (vert_segments == NULL) {
        CGImageRelease (game_image) ;
        free (raw_image_file_data) ;
        freeLayout (layout, layout_nb_columns * layout_nb_lines) ;
        freePieces (pieces, nb_columns_of_pieces * nb_lines_of_pieces) ;
        displayMessage ("Error", "Not enough memory.", AC_Critical) ;
        return (NULL) ;
      }
      
      vert_spline =
        buildSpline
          (1, vert_segments, NB_UNIT_EDGE_POINTS, &spline_top_left,
           &spline_bottom_right) ;
      if (vert_spline == NULL) {
        free (vert_segments) ;
        CGImageRelease (game_image) ;
        free (raw_image_file_data) ;
        freeLayout (layout, layout_nb_columns * layout_nb_lines) ;
        freePieces (pieces, nb_columns_of_pieces * nb_lines_of_pieces) ;
        displayMessage ("Error", "Not enough memory.", AC_Critical) ;
        return (NULL) ;
      }
      
      /* Release the segments, they are useless now. */
      free (vert_segments) ;

      /* Initialise the west of the current cell. */
      /* pieces [x][y]. */
      index = (y * nb_columns_of_pieces) + x ;
      pieces[index].free_edges[WEST]->spline = vert_spline ;
      pieces[index].delta_center_to_top_left.x =
        min
          (pieces[index].delta_center_to_top_left.x,
           (spline_top_left.x - pieces[index].center.x)) ;
      pieces[index].delta_center_to_top_left.y =
        min
          (pieces[index].delta_center_to_top_left.y,
           (spline_top_left.y - pieces[index].center.y)) ;
      pieces[index].delta_center_to_bottom_right.x =
        max
         (pieces[index].delta_center_to_bottom_right.x,
          (spline_bottom_right.x - pieces[index].center.x)) ;
      pieces[index].delta_center_to_bottom_right.y =
        max
          (pieces[index].delta_center_to_bottom_right.y,
           (spline_bottom_right.y - pieces[index].center.y)) ;
      
      /* Init of the east of the previous x cell. */
      if (x > 0) {
        /* piece[x-1][y]. */
        index = (y * nb_columns_of_pieces) + (x - 1) ;
        pieces[index].free_edges[EAST]->spline =
          build_opposite_spline (/* shift_up: */ 0, vert_spline) ;
        pieces[index].delta_center_to_top_left.x =
          min
            (pieces[index].delta_center_to_top_left.x,
             (spline_top_left.x - pieces[index].center.x)) ;
        pieces[index].delta_center_to_top_left.y =
          min
            (pieces[index].delta_center_to_top_left.y,
             (spline_top_left.y - pieces[index].center.y)) ;

        pieces[index].delta_center_to_bottom_right.x =
          max
            (pieces[index].delta_center_to_bottom_right.x,
             (spline_bottom_right.x - pieces[index].center.x)) ;
        pieces[index].delta_center_to_bottom_right.y =
          max
            (pieces[index].delta_center_to_bottom_right.y,
             (spline_bottom_right.y - pieces[index].center.y)) ;
      }    /* End of if (x > 0). */
    }      /* End of for (x = 1; x < nb_columns; x++). */
  }        /* End of for (y = 0; y < nb_lines; y++). */

  /* Initialize the sorted double linked lists of increasing x. */
  prev = NULL ;
  for (x = 0; x < nb_columns_of_pieces; x++) {
    for (y = 0; y < nb_lines_of_pieces; y++) {
      /* Processing north edge of the current cell (pieces[x][y]). */
      tmp_edge = pieces[(y * nb_columns_of_pieces) + x].free_edges[NORTH] ;
      if (prev != NULL) insertAfterInXList (tmp_edge, prev) ;
      prev = tmp_edge ;
      /* Processing east edge of the left cell if some (pieces[x-1][y]). */
      if (x > 0) {
        tmp_edge =
          pieces[(y * nb_columns_of_pieces) + (x - 1)].free_edges[EAST] ;
        if (prev != NULL) insertAfterInXList (tmp_edge, prev) ;
        prev = tmp_edge ;
      }
      /* Processing the west edge of the current cell (pieces[x][y]). */
      tmp_edge = pieces[(y * nb_columns_of_pieces) + x].free_edges[WEST] ;
      if (prev != NULL) insertAfterInXList (tmp_edge, prev) ;
      prev = tmp_edge ;
      /* Processing the south edge of the current cell (pieces[x][y]). */
      tmp_edge = pieces[(y * nb_columns_of_pieces) + x].free_edges[SOUTH] ;
      if (prev != NULL) insertAfterInXList (tmp_edge, prev) ;
      prev = tmp_edge ;
    }   /* End of for (y = 0; y < nb_lines; y++). */
  }     /* End of for (x = 0; x < nb_columns; x++). */

  /* Processing the east edges of the last column (pieces[nb_columns-1][y]). */
  for (y = 0; y < nb_lines_of_pieces; y++) {
    tmp_edge =
      pieces[(y * nb_columns_of_pieces) + (nb_columns_of_pieces - 1)].
        free_edges[EAST] ;
    if (prev != NULL) insertAfterInXList (tmp_edge, prev) ;
    prev =tmp_edge ;
  }

  /* Initialize the sorted double linked lists of increasing y. */
  prev = NULL ;
  for (y = 0; y < nb_lines_of_pieces; y++) {
    for (x = 0 ; x < nb_columns_of_pieces; x++) {
      /* Processing south edge of the above cell if some (pieces[x][y-1]). */
      if (y > 0) {
        tmp_edge =
          pieces[((y - 1) * nb_columns_of_pieces) + x].free_edges[SOUTH] ;
        if (prev != NULL) InsertAfterInYList (tmp_edge, prev) ;
        prev = tmp_edge ;
      }
      /* Processing north edge of the current cell (pieces[x][y]). */
      tmp_edge = pieces[(y * nb_columns_of_pieces) + x].free_edges[NORTH] ;
      if (prev != NULL) InsertAfterInYList (tmp_edge, prev) ;
      prev = tmp_edge ;
      /* Processing the west edge of the current cell (pieces[x][y]). */
      tmp_edge = pieces[(y * nb_columns_of_pieces) + x].free_edges[WEST] ;
      if (prev != NULL) InsertAfterInYList (tmp_edge, prev) ;
      prev = tmp_edge ;
      /* Processing east edge of the current cell (pieces[x][y]). */
      tmp_edge = pieces[(y * nb_columns_of_pieces) + x].free_edges[EAST] ;
      if (prev != NULL) InsertAfterInYList (tmp_edge, prev) ;
      prev = tmp_edge ;
    }   /* End of for (x = 0 ; x < nb_columns; x++). */
  }     /* End of for (y = 0; y <nb_lines; y++). */

  /* Processing the south edges of the last line (pieces[x][nb_lines - 1]). */
  for (x = 0; x < nb_columns_of_pieces; x++) {
    tmp_edge =
      pieces[((nb_lines_of_pieces - 1) * nb_columns_of_pieces) + x].
      free_edges[SOUTH] ;
    if (prev != NULL) InsertAfterInYList (tmp_edge, prev) ;
    prev = tmp_edge ;
  }

  /* Create the game structure. */
  game = malloc (sizeof (struct game)) ;
  if (game == NULL) {
    CGImageRelease (game_image) ;
    free (raw_image_file_data) ;
    freeLayout (layout, layout_nb_columns * layout_nb_lines) ;
    freePieces (pieces, nb_columns_of_pieces * nb_lines_of_pieces) ;
    displayMessage ("Error", "Not enough memory.", AC_Critical) ;
    return (NULL) ;    /* Memory leak. */
  }
  
  memset (game, 0, sizeof (struct game)) ;
  game->game_won = FALSE ;
  game->image_width = image_width ;
  game->image_height = image_height ;
  game->original_image_scale_factor = scale_factor ;
  game->piece_width = piece_width ;
  game->piece_height = piece_height ;
  game->layout_nb_lines = layout_nb_lines ;
  game->layout_nb_columns = layout_nb_columns ;
  game->layout = layout ;
  game->pieces = pieces ;
  game->nb_pieces = nb_columns_of_pieces * nb_lines_of_pieces ;
  /* By construction, the first edge is north of the first piece ([0][0]). */
  game->x_sorted_edges_head = pieces[0].free_edges[NORTH] ;
  /* By construction, the last edge is east of the last piece
     ([nb_columns - 1][nb_lines - 1]) . */
  game->x_sorted_edges_tail =
    pieces
     [((nb_lines_of_pieces - 1) * nb_columns_of_pieces) +
      (nb_columns_of_pieces - 1)].free_edges[EAST] ;
  /* By construction, the first edge is north of the first piece ([0][0]). */
  game->y_sorted_edges_head = pieces[0].free_edges[NORTH] ;
  /* By construction, the last edge is south of the last piece
     ([nb_columns - 1][nb_lines - 1]). */
  game->y_sorted_edges_tail =
    pieces
      [((nb_lines_of_pieces - 1) * nb_columns_of_pieces) +
       (nb_columns_of_pieces - 1)].free_edges[SOUTH] ;
  game->game_image = game_image ;
  game->image_file_size = raw_image_file_size ;
  game->image_file_data = raw_image_file_data ;
  
  /* Now bounding-boxes are computed, and we have a game structure
     we can record the pieces in the layout and by the way texture
     map them. */
  for (x = 0; x < nb_columns_of_pieces; x++) {
    for (y = 0; y < nb_lines_of_pieces; y++) {
      /* pieces[x][y]. */
      struct piece *piece = &pieces[(y * nb_columns_of_pieces) + x] ;
      
      if (! insertPieceInLayout (game, piece)) {
        displayMessage ("Error", "Not enough memory.", AC_Critical) ;
        freeGame (game) ;
        return (NULL) ;
      }
      
      /* Since the game is not yet shuffled, pieces are still positionned
         exactly matching their position in the game picture. So we can directly
         use the center position that is recorded in the piece itself. */
      if (! initializePieceTexture (game->game_image, piece, &piece->center)) {
        displayMessage ("Error", "Unable to initialize textures.", AC_Critical) ;
        freeGame (game) ;
        return (NULL) ;   /* MEMLEAK */
      }
    }    /* End of for (y = 0; y < nb_lines; y++). */
  }      /* End of for (x = 0; x < nb_columns; x++). */

  /* Finally, return the game structure. */
  return (game) ;
}

