/*
  game_load.c
  jigplay-macos

  Created by François Pessaux on 30/09/10.
  Copyright 2010. All rights reserved.
*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#ifdef __JIGPLAY_IOS
#include <CoreGraphics/CGBitmapContext.h>
#endif

#include "misc.h"
#include "game_load.h"
#include "game_alloc.h"
#include "game_free.h"
#include "alerts_n_errors.h"
#include "texturemap.h"


/** Local temporary structure used to rebuild connected sets circumvening the
    fact that when loadign them from the saved game, pieces are not yet
    loaded. */
struct pre_connected_set {
  struct connected_set *partial_connected_set ;
  /* Number of pieces members of the pre-connected-set, hence of the related
     future connected-set. */
  int nb_members ;
  /* An array containing the ids of the pieces members of this
     pre-connected-set, hence of the related future connected-set. */
  int *members_ids ;
} ;



struct pre_edge {
  struct edge *partial_edge ;
  int future_edge_part_of_piece_id ;
  int future_edge_x_sorted_prev_id ;
  int future_edge_x_sorted_next_id ;
  int future_edge_y_sorted_prev_id ;
  int future_edge_y_sorted_next_id ;
};



/** Remember that the array of pieces is expected to be indexed by the pieces
    ids. */
static struct layout_cell* loadLayout (FILE *in, int nb_cells,
                                       struct piece *pieces)
{
  int i, read ;
  struct layout_cell *layout = NULL ;
  
  /* Allocate the layout itself. */
  layout = malloc (nb_cells * sizeof (struct layout_cell)) ;
  if (layout == NULL) return (NULL) ;
  memset (layout, 0, sizeof (nb_cells * sizeof (struct layout_cell))) ;
  
  /* Load the list of pieces ids of each layout cell. Set the redraw flag
     to false because we will anyway redraw the whole screen at startup. */
  for (i = 0; i < nb_cells; i++) {
    struct pieces_list *list_head = NULL ;
    int nb_pieces, j ;
    int *ids ;
    
    /* Read the number of pieces in the list. */
    read = fread (&nb_pieces, sizeof (int), 1, in) ;
    if (read != 1) {
      freeLayout (layout, nb_cells) ;
      return (NULL) ;
    }
    
    /* Now, read all the ids in one shot. */
    ids = malloc (nb_pieces * sizeof (int)) ;
    if (ids == NULL) {
      freeLayout (layout, nb_cells) ;
      displayMessage ("Error", "Not enough memory.", AC_Critical) ;
      return (NULL) ;
    }
    
    read = fread (ids, sizeof (int), nb_pieces, in) ;
    if (read != nb_pieces) {
      freeLayout (layout, nb_cells) ;
      free (ids) ;
      return (NULL) ;
    }
    
    /* Rebuild the list starting by the end of the array to be able to add in
       head, which is more efficient. */
    for (j = nb_pieces - 1; j >= 0; j--) {
      struct pieces_list *new_cell = malloc (sizeof (struct pieces_list)) ;
      if (new_cell == NULL) {
        free (ids) ;
        /* Hook the head of the list built upon this point, so that freeing
           the layout structure will free the cells we allocated until we
           failed here. */
        layout[i].hosted_pieces = list_head ;
        freeLayout (layout, nb_cells) ;
        displayMessage ("Error", "Not enough memory.", AC_Critical) ;
        return (NULL) ;
      }     /* End of if (new_cell == NULL). */
      
      new_cell->piece = &pieces[ids[j]] ;
      /* Insert in head of the list. */
      new_cell->next = list_head ;
      list_head = new_cell ;
    }    /* End of for (j = nb_cells - 1; j >= 0; j--). */
    
    free (ids) ;
    
    layout[i].to_redraw = FALSE ;
    layout[i].hosted_pieces = list_head ;
  }     /* End of for (i = 0; i < nb_cells; i++). */
  
  return (layout) ;
}



static struct spline* loadSpline (FILE *in)
{
  int read ;

  struct spline *spline = malloc (sizeof (struct spline)) ;
  if (spline == NULL) {
    displayMessage ("Error", "Not enough memory.", AC_Critical) ;
    return (NULL) ;
  }
  
  /* First, read the number of points. */
  read = fread (&spline->nb_points, sizeof (int), 1, in) ;
  if (read != 1) {
    free (spline) ;
    return (NULL) ;
  }
  
  spline->vertices = malloc (spline->nb_points * sizeof (struct point)) ;
  if (spline->vertices == NULL) {
    free (spline) ;
    displayMessage ("Error", "Not enough memory.", AC_Critical) ;
    return (NULL) ;
  }
  
  read = fread (spline->vertices, sizeof (struct point), spline->nb_points, in) ;
  if (read != spline->nb_points) {
    free (spline->vertices) ;
    free (spline) ;
    spline = NULL ;
  }
  
  return (spline) ;
}



/** Function only used to release the memory occupied by pre-edges in case of
    error, i.e. in case where the complete setup of pre-edges failed, hence
    everything allocated must be released. Especially, this will also free the
    edges structures allocated and partially initialized.
    This function must hence not obviously be called to free the pre-edges
    if no error was detected because this would also free the edges that should
    be shared later in the game structure. */
static void freePreEdgesOnError (struct pre_edge *pre_edges, int nb_pre_edges)
{
  int i ;
  
  if (pre_edges == NULL) return ;
  for (i = 0; i < nb_pre_edges; i++) {
    if (pre_edges[i].partial_edge != NULL) free (pre_edges[i].partial_edge) ;
  }
  
  free (pre_edges) ;
}



/** Create an array containing the pre-edges (that have a pointer on the
    allocated edge) from the saved edges description. These pre-edges contain
    partially initialized edges, in particular their links in the
    double-linked X/Y sorted lists are not set and the piece the edge is part
    of is not set. These pre-edges also contain material (ids) to be able later
    to fully initialize the edges pointers.
    The aim of this function is to pre-create the edges, store them in an
    array indexed by edges ids in order to be able to have all the addresses
    of edges availables to later create pieces and link pieces in the double
    linked X/Y sorted lists and in the pieces.
    In fact, since we have mutualm references between edges and between edges
    and pieces, we can't rebuild the data-structures in one pass. */
static struct pre_edge* loadAllEdgesAndMakePreEdges (FILE *in, int nb_edges)
{
  int i, read ;
  int status = TRUE ;
  struct pre_edge *all_pre_edges = NULL ;
  
  /* Allocate the array to remind the edges pointers. Once pieces and edges are
     fully initialized, this array can be freed (but not the edges the
     pre-edges have pointers on, of course, since they are retained in the
     pieces and other edges data-structures. */
  all_pre_edges = malloc (nb_edges * sizeof (struct pre_edge)) ;
  if (all_pre_edges == NULL) {
    displayMessage ("Error", "Not enough memory", AC_Critical) ;
    return (NULL) ;
  }
  memset (all_pre_edges, 0, nb_edges * sizeof (struct pre_edge)) ; 
  
  for (i = 0; i < nb_edges; i++) {
    struct edge *edge ;
    struct pre_edge *curr_pre_edge ;
    
    edge = malloc (sizeof (struct edge)) ;
    if (edge == NULL) {
      displayMessage ("Error", "Not enough memory.", AC_Critical) ;
      status = FALSE ;
      goto quick_end ;
    }

    /* Read the edge id. */
    read = fread (&edge->id, sizeof (int), 1, in) ;
    if (read != 1) { status = FALSE ; goto quick_end ; }
    
    /* Store the edge in the array pre-edge cell at the index corresponding to
       its id and eemind the address of the pre-edge we are building. */
    all_pre_edges[edge->id].partial_edge = edge ;
    curr_pre_edge = &all_pre_edges[edge->id] ;

    /* Edge activity. */
    read = fread (&edge->active, sizeof (int), 1, in) ;
    if (read != 1) { status = FALSE ; goto quick_end ; }
    /* Point p0. */
    read = fread (&edge->p0, sizeof (struct point), 1, in) ;
    if (read != 1) { status = FALSE ; goto quick_end ; }
    /* Point p0 anchor id. */
    read = fread (&edge->p0_anchor_id, sizeof (int), 1, in) ;
    if (read != 1) { status = FALSE ; goto quick_end ; }
    /* Point p1. */
    read = fread (&edge->p1, sizeof (struct point), 1, in) ;
    if (read != 1) { status = FALSE ; goto quick_end ; }
    /* Point p1 anchor id. */
    read = fread (&edge->p1_anchor_id, sizeof (int), 1, in) ;
    if (read != 1) { status = FALSE ; goto quick_end ; }

    /* The spline. */
    edge->spline = loadSpline (in) ;
    if (edge->spline == NULL) { status = FALSE ; goto quick_end ; }
    
    /* The id of the piece this edge belongs to. */
    read =
      fread
        (&curr_pre_edge->future_edge_part_of_piece_id, sizeof (int), 1, in) ;
    if (read != 1) { status = FALSE ; goto quick_end ; }
    /* The ids of our next/prev edges in the X/Y sorted lists. */
    read =
      fread
        (&curr_pre_edge->future_edge_x_sorted_prev_id, sizeof (int), 1, in) ;
    if (read != 1) { status = FALSE ; goto quick_end ; }
    read =
      fread
        (&curr_pre_edge->future_edge_x_sorted_next_id, sizeof (int), 1, in) ;
    if (read != 1) { status = FALSE ; goto quick_end ; }
    read =
      fread
        (&curr_pre_edge->future_edge_y_sorted_prev_id, sizeof (int), 1, in) ;
    if (read != 1) { status = FALSE ; goto quick_end ; }
    read =
      fread
        (&curr_pre_edge->future_edge_y_sorted_next_id, sizeof (int), 1, in) ;
    if (read != 1) { status = FALSE ; goto quick_end ; }
  }    /* End of for (i = 0; i < nb_edges; i++). */
  
quick_end:
  if (status == FALSE) {
    freePreEdgesOnError (all_pre_edges, nb_edges) ;
    all_pre_edges = NULL ;
  }
  
  return (all_pre_edges) ;
}



/** Remember that the array of pieces is expected to be indexed by the pieces
    ids. */
static bool_t finalizeConnectedSets
    (struct pre_connected_set *pre_connected_sets,
     int nb_connected_sets, struct piece *pieces, int nb_pieces)
{
  int i ;
  
  /* Now the pieces are well-ordered in the array, we can finalize the
     connected-sets one by one. */
  for (i = 0; i < nb_connected_sets; i++) {
    int j ;
    struct pieces_list *list_head = NULL ;
    
    /* Rebuild the list starting by the end of the array to be able to add in
       head, which is more efficient. */
    for (j = pre_connected_sets[i].nb_members - 1; j >= 0; j--) {
      struct pieces_list *new_cell = malloc (sizeof (struct pieces_list)) ;
      if (new_cell == NULL) {
        displayMessage ("Error", "Not enough memory", AC_Critical) ;
        /* Hook the list we succeeded to build upon here in the connected
           set so that memory allocated until we failed here will be freed
           anyway. */
        pre_connected_sets[i].partial_connected_set->members = list_head ;
        return (FALSE) ;
      }

      new_cell->piece = &pieces[pre_connected_sets[i].members_ids[j]] ;
      /* Insert in head of the list. */
      new_cell->next = list_head ;
      list_head = new_cell ;
    }    /* End of for (j = pre_connected_sets[i].nb_members - 1; ...). */
    
    /* Finally, record the head of the list. */
    pre_connected_sets[i].partial_connected_set->members = list_head ;
  }   /* End of for (i = 0; i < nb_connected_sets; i++). */
  
  return (TRUE) ;
}



/** Terminates the edges initialization by updating the double-linked X/Y
    sorted lists. Until all the edges were created, we didn't have all the
    pointers on the edges. When this function is called, we assume we have all
    these pointers. Hence, we can finally make the pointers pointing at the
    rigth places in these lists in the edges.
    The array of pre-edges we receive is expected to be indexed by the edges
    ids. Hence when searching an edge with a certain id, we will find it in
    the array at the index equal to this id. */
static void finalizeEdgesLinkedLists (struct pre_edge *pre_edges, int nb_edges)
{
  int i ;
  
  for (i = 0; i < nb_edges; i++) {
    struct pre_edge *pre_edge = &pre_edges[i] ;
    struct edge *edge = pre_edge->partial_edge ;
    
    /* Deal with the x_sorted_prev. */
    edge->x_sorted_prev =
      ((pre_edge->future_edge_x_sorted_prev_id == -1) ?
       NULL : pre_edges[pre_edge->future_edge_x_sorted_prev_id].partial_edge) ;
    /* Deal with the x_sorted_next. */
    edge->x_sorted_next =
      ((pre_edge->future_edge_x_sorted_next_id == -1) ?
       NULL : pre_edges[pre_edge->future_edge_x_sorted_next_id].partial_edge) ;
    /* Deal with the y_sorted_prev. */
    edge->y_sorted_prev =
      ((pre_edge->future_edge_y_sorted_prev_id == -1) ?
       NULL : pre_edges[pre_edge->future_edge_y_sorted_prev_id].partial_edge) ;
    /* Deal with the y_sorted_next. */
    edge->y_sorted_next =
      ((pre_edge->future_edge_y_sorted_next_id == -1) ?
       NULL : pre_edges[pre_edge->future_edge_y_sorted_next_id].partial_edge) ;
  }    /* End of for (i = 0; i < nb_edges; i++). */
}



/** Function only used to release the memory occupied by pre-connected-sets in
    case of error, i.e. in case where the complete setup of pre-connected-sets
    failed, hence everything allocated must be released. Especially, this will
    also free the connected-sets structures allocated and partially initialized.
    This function must hence not obviously be called to free the
    pre-connected-sets if no error was detected because this would also free the
    connected-sets that should be shared later in the game structure. */
static void freePreConnecteSetsOnError
    (struct pre_connected_set *pre_connected_sets,
     int nb_pre_connected_sets)
{
  int i ;
  
  if (pre_connected_sets == NULL) return ;
  for (i = 0; i < nb_pre_connected_sets; i++) {
    if (pre_connected_sets[i].members_ids != NULL)
      free (pre_connected_sets[i].members_ids) ;
  }
  
  free (pre_connected_sets) ;
}



/** x_head, x_tail, y_head and y_tail are out-mode parameters. */
struct piece* loadPieces (FILE *in, int nb_pieces,
                          int x_head_id, int x_tail_id,
                          int y_head_id, int y_tail_id,
                          struct edge **x_head, struct edge **x_tail,
                          struct edge **y_head, struct edge **y_tail)
{
  int nb_connected_sets, read, i ;
  int status = TRUE ;
  struct pre_edge *all_pre_edges = NULL ;
  struct piece *pieces = NULL ;
  struct pre_connected_set *pre_connected_sets = NULL ;
  
  /* We first read the number of saved connected-sets. */
  read = fread (&nb_connected_sets, sizeof (int), 1, in) ;
  if (read != 1) { status = FALSE ; goto quick_end ; }

  pre_connected_sets =
    malloc (nb_connected_sets * sizeof (struct pre_connected_set)) ;
  if (pre_connected_sets == NULL) {
    displayMessage ("Error", "Not enough memory.", AC_Critical) ;
    return (NULL) ;
  }
  memset
    (pre_connected_sets, 0,
     nb_connected_sets * sizeof (struct pre_connected_set)) ;
  
  /* Now read the saved connected-sets to make pre-connected-sets. */
  for (i = 0; i < nb_connected_sets; i++) {
    /* The number of pieces members of this connected-set. */
    read = fread (&pre_connected_sets[i].nb_members, sizeof (int), 1, in) ;
    if (read != 1) { status = FALSE ; goto quick_end ; }
    pre_connected_sets[i].members_ids =
      malloc (pre_connected_sets[i].nb_members * sizeof (int)) ;
    if (pre_connected_sets[i].members_ids == NULL) {
      displayMessage ("Error", "Not enough memory.", AC_Critical) ;
      status = FALSE ;
      goto quick_end ;
    }
    
    /* Create the related real connected-set structure. */
    pre_connected_sets[i].partial_connected_set =
      malloc (sizeof (struct connected_set)) ;
    if (pre_connected_sets[i].partial_connected_set == NULL) {
      displayMessage ("Error", "Not enough memory.", AC_Critical) ;
      status = FALSE ;
      goto quick_end ;
    }
    
    /* Load the refcount directly in the connected-set structure. */
    read =
      fread
        (&pre_connected_sets[i].partial_connected_set->refcount, sizeof (int),
         1, in) ;
    if (read != 1) { status = FALSE ; goto quick_end ; }
    
    /* Load the ids. */
    read =
      fread
        (pre_connected_sets[i].members_ids, sizeof (int),
         pre_connected_sets[i].nb_members, in) ;
    if (read != pre_connected_sets[i].nb_members) {
      status = FALSE ; goto quick_end ;
    }
  }
 
  /* Ok, right, connected-sets information is loaded, future connected-sets
     structures are allocated, pre-connected-sets are built, we will have later
     to complete connected-sets structures by making lists of pieces whose
     indices are in the pre-connected-sets structures. But we will be able to
     do thing only once pieces are loaded and created.
     Now, address loading edges. */
  all_pre_edges = loadAllEdgesAndMakePreEdges (in, nb_pieces * 4) ;
  if (all_pre_edges == NULL) { status = FALSE ; goto quick_end ; }
  
  /* Now, address loading pieces information. */
  pieces = malloc (nb_pieces * sizeof (struct piece)) ;
  if (pieces == NULL) {
    displayMessage ("Error", "Not enough memory.", AC_Critical) ;
    status = FALSE ;
    goto quick_end ;
  }
  memset (pieces, 0, nb_pieces * sizeof (struct piece)) ;
  
  for (i = 0; i < nb_pieces; i++) {
    struct piece *piece = &pieces[i] ;
    int tmp, j ;
    
    /* Read the id. */
    read = fread (&piece->piece_debug_id, sizeof (int), 1, in) ;
    if (read != 1) { status = FALSE ; goto quick_end ; }
    /* The center. */
    read = fread (&piece->center, sizeof (struct point), 1, in) ;
    if (read != 1) { status = FALSE ; goto quick_end ; }
    /* The bounding-box. */
    read =
      fread (&piece->delta_center_to_top_left, sizeof (struct point), 1, in) ;
    if (read != 1) { status = FALSE ; goto quick_end ; }
    read =
      fread (&piece->delta_center_to_bottom_right, sizeof (struct point), 1, in) ;
    if (read != 1) { status = FALSE ; goto quick_end ; }

    /* The edges. We assing the edges in the piece and by the way, we update the
       field part_of_piece of the related edge. */
    for (j = 0; j < 4; j++) {
      read = fread (&tmp, sizeof (int), 1, in) ;
      if (read != 1) { status = FALSE ; goto quick_end ; }
      piece->free_edges[j] = all_pre_edges[tmp].partial_edge ;
      piece->free_edges[j]->part_of_piece = piece ;
    }

    /* Recover the connected-set this piece belongs to. We read the index of
       the connected-set. */
    read = fread (&tmp, sizeof (int), 1, in) ;
    if (read != 1) { status = FALSE ; goto quick_end ; }
    piece->connected_pieces = pre_connected_sets[tmp].partial_connected_set ;
    
    /* The texture will be rebuilt later, once the picture is loaded. */
  }     /* End of for (i = 0; i < nb_pieces; i++). */

  /* Now, terminate the edges initialization by setting the right pointers in
     the double-linked X/Y sorted lists inside the pieces. Since the field
     part_of_piece of each edge has been initialized while creating the pieces,
     after having initialized the linked lists, edges should be finally fully
     initialized. */
  finalizeEdgesLinkedLists (all_pre_edges, nb_pieces * 4) ;
 
  /* Set by side effect the out-mode parameters telling what are the pointer on
     heads and tails of the double-linked X/ sorted list. */
  *x_head = ((x_head_id == -1) ? NULL : all_pre_edges[x_head_id].partial_edge) ;
  *x_tail = ((x_tail_id == -1) ? NULL : all_pre_edges[x_tail_id].partial_edge) ;
  *y_head = ((y_head_id == -1) ? NULL : all_pre_edges[y_head_id].partial_edge) ;
  *y_tail = ((y_tail_id == -1) ? NULL : all_pre_edges[y_tail_id].partial_edge) ;

  /* We don't need anymore the array of pre-edges, so free it. */
  free (all_pre_edges) ;
  all_pre_edges = NULL ;
  
  /* Now the pieces are created set the pointers to pieces in the
     connected-sets. */
  finalizeConnectedSets
    (pre_connected_sets, nb_connected_sets, pieces, nb_pieces) ;

  /* We dont need anymore the pre-connected-sets and the array storing the
     pointer to the connected-sets, so free them. */
  free (pre_connected_sets) ;
  pre_connected_sets = NULL ;

quick_end:
  if (status == FALSE) {
    if (pieces != NULL) {
      /* If pieces were allocated, then we don't call the routines to free
         memory on error on pre-edges and pre-connected-sets because the
         pointers to the edges and connected-sets of these structure may
         already be set in the piece structure and we would free them twice.
         This has the drawback that some memory may leak in case where the
         pieces were not totally created since some edges and connected-sets
         not already hooked in a piece structure won't be freed because no
         piece make reference to them. Bof... for so much memory, give up ! */
      freePieces (pieces, nb_pieces) ;
      if (pre_connected_sets != NULL) free (pre_connected_sets) ;
      if (all_pre_edges != NULL) free (all_pre_edges) ;
      pieces = NULL ;
    }
    else {
      /* Pieces are already NULL, so no connected-set nor edge pointers are
         already hooked in pieces structures, so we can safely free them
         with the error recovery routines. */
      freePreConnecteSetsOnError (pre_connected_sets, nb_connected_sets) ;
      freePreEdgesOnError (all_pre_edges, nb_pieces * 4) ;
    }
  }
  
  return (pieces) ;
}



static CGImageRef rebuildImageFromSavedBytes (unsigned char *data,
                                              int size, float scale_factor)
{
  CGDataProviderRef provider = NULL ;
  CGImageRef image = NULL ;
  CGImageAlphaInfo alphaInfo ;
  const char *tmp_fname ;
  FILE *tmp_file ;
  
  tmp_fname = tmpnam (NULL) ;
  tmp_file = fopen (tmp_fname, "wb") ;
  if (tmp_file == NULL) return (NULL) ;
  fwrite (data, sizeof (unsigned char), size, tmp_file) ;
  fclose (tmp_file) ;
  
  provider = CGDataProviderCreateWithFilename (tmp_fname) ;
  if (provider == NULL) {
    unlink (tmp_fname) ;
    return (NULL) ;
  }
  
  /* Try to detect the PNG header. */
  if ((data[0] == 0x89) && (data[1] == 0x50) &&
      (data[2] == 0x4E) && (data[3] == 0x47)) {
    image =
      CGImageCreateWithPNGDataProvider
        (provider, NULL, true, kCGRenderingIntentDefault) ;
  }
  else {
    image =
      CGImageCreateWithJPEGDataProvider
        (provider, NULL, true, kCGRenderingIntentDefault) ;
  }
  
  CGDataProviderRelease (provider) ;
  unlink (tmp_fname) ;
  
  if (image == NULL) return (NULL) ;
  
  if (scale_factor > 0) {
    CGImageRef tmp = image ;
    /* No need to check if we need to add transparency, the scaling routine
       always add it by the way because this doesn't cost it more. */
    image = makeScaledCopyOfImageAndAddAlphaChannel (tmp, scale_factor) ;
    CGImageRelease (tmp) ;
  }
  else {
    /* If the image doesn't have any transparency channel, add one. */
    alphaInfo = CGImageGetAlphaInfo (image) ;
    if ((alphaInfo == kCGImageAlphaNone) ||
        (alphaInfo == kCGImageAlphaNoneSkipLast) ||
        (alphaInfo == kCGImageAlphaNoneSkipFirst)) {
      CGImageRef tmp = image ;
      image = copyImageAndAddAlphaChannel (tmp) ;
      CGImageRelease (tmp) ;
    }
  }
  
  return (image) ;  
}



struct game *loadGame (const char *file_name)
{
  FILE *in ;
  int tmp, read ;
  int status = TRUE ;
  int x_head_id, x_tail_id, y_head_id, y_tail_id ;
  struct edge *x_head, *x_tail, *y_head, *y_tail ;
  struct game *game ;
  int nb_columns_of_pieces, nb_lines_of_pieces, x, y ;
  
  in = fopen (file_name, "rb") ;
  if (in == NULL) {
    displayMessage ("Error", "Unable to open file.", AC_Critical) ;
  return (NULL) ;
  }
  
  /* First, check the magic number. */
  read = fread (&tmp, sizeof (int), 1, in) ;
  if (read != 1) { status = FALSE ; goto quick_end ; }
  if (tmp != SAVEMAGICNUMBER) {
    fclose (in) ;
    displayMessage
      ("Error", "File doesn't seem to be a saved game.", AC_Critical) ;
    return (NULL) ;
  }
  
  /* Allocate memory for the main game structure. */
  game = malloc (sizeof (struct game)) ;
  if (game == NULL) {
    fclose (in) ;
    displayMessage ("Error", "Not enough memory.", AC_Critical) ;
    return (NULL) ;
  }
  
  memset (game, 0, sizeof (struct game)) ;
  game->game_won = FALSE ;
  
  /* Read the miscellaneous scalars. */
  read = fread (&game->image_width, sizeof (int), 1, in) ;
  if (read != 1) { status = FALSE ; goto quick_end ; }
  read = fread (&game->image_height, sizeof (int), 1, in) ;
  if (read != 1) { status = FALSE ; goto quick_end ; }
  read = fread (&game->original_image_scale_factor, sizeof (float), 1, in) ;
  if (read != 1) { status = FALSE ; goto quick_end ; }
  read = fread (&game->piece_width, sizeof (int), 1, in) ;
  if (read != 1) { status = FALSE ; goto quick_end ; }
  read = fread (&game->piece_height, sizeof (int), 1, in) ;
  if (read != 1) { status = FALSE ; goto quick_end ; }
  read = fread (&game->layout_nb_columns, sizeof (int), 1, in) ;
  if (read != 1) { status = FALSE ; goto quick_end ; }
  read = fread (&game->layout_nb_lines, sizeof (int), 1, in) ;
  if (read != 1) { status = FALSE ; goto quick_end ; }
  read = fread (&game->nb_pieces, sizeof (int), 1, in) ;
  if (read != 1) { status = FALSE ; goto quick_end ; }
 
  /* The ids of the heads of the x and y sorted lists of edges or -1 if NULL. */
  read = fread (&x_head_id, sizeof (int), 1, in) ;
  if (read != 1) { status = FALSE ; goto quick_end ; }
  read = fread (&x_tail_id, sizeof (int), 1, in) ;
  if (read != 1) { status = FALSE ; goto quick_end ; }
  read = fread (&y_head_id, sizeof (int), 1, in) ;
  if (read != 1) { status = FALSE ; goto quick_end ; }
  read = fread (&y_tail_id, sizeof (int), 1, in) ;
  if (read != 1) { status = FALSE ; goto quick_end ; }
  
  /* Now, it's time to load the pieces. By the way, recover by side effect the
     addresses of the edges heads and tails of the double-linked X/Y sorted
     lists. */
  game->pieces =
    loadPieces
      (in, game->nb_pieces, x_head_id, x_tail_id, y_head_id, y_tail_id,
       &x_head, &x_tail, &y_head, &y_tail) ;
  if (game->pieces == NULL) { status = FALSE ; goto quick_end ; }
  
  game->x_sorted_edges_head = x_head ;
  game->x_sorted_edges_tail = x_tail ;
  game->y_sorted_edges_head = y_head ;
  game->y_sorted_edges_tail = y_tail ;

  /* Load the layout. */
  game->layout =
    loadLayout
      (in, game->layout_nb_columns * game->layout_nb_lines, game->pieces) ;
  if (game->layout == NULL) { status = FALSE ; goto quick_end ; }
  
  /* The image file data. First, we read it. */
  read = fread (&game->image_file_size, sizeof (int), 1, in) ;
  if (read != 1) { status = FALSE ; goto quick_end ; }
  game->image_file_data =
    malloc (game->image_file_size * sizeof (unsigned char)) ;
  if (game->image_file_data == NULL) {
    displayMessage ("Error", "Not enough memory.", AC_Critical) ;
    status = FALSE ;
    goto quick_end ;
  }
  read =
    fread
      (game->image_file_data, sizeof (unsigned char), game->image_file_size,
       in) ;
  if (read != game->image_file_size) { status = FALSE ; goto quick_end ; }
  
  /* Now, rebuild a CGImage ref from this content. */
  game->game_image =
    rebuildImageFromSavedBytes
      (game->image_file_data, game->image_file_size,
       game->original_image_scale_factor) ;
  if (game->game_image == NULL) { status = FALSE ; goto quick_end ; }
  
  /* We must now texture the pieces. */
  nb_columns_of_pieces = game->image_width / game->piece_width ;
  nb_lines_of_pieces = game->image_height / game->piece_height ;
  for (x = 0; x < nb_columns_of_pieces; x++) {
    for (y = 0; y < nb_lines_of_pieces; y++) {
      struct point initial_center_position ;
    
      /* Since the game was loaded, pieces moved compared to their position in a
         new game before shuffling. So, rebuild the position of the center. */
      initial_center_position.x =
        x * game->piece_width + (game->piece_width / 2) ;
      initial_center_position.y =
        y * game->piece_height + (game->piece_height / 2) ;
      if (! initializePieceTexture
              (game->game_image, &game->pieces[(y * nb_columns_of_pieces) + x],
               &initial_center_position)) {
        status = FALSE ;
        goto quick_end ;
      }
    }    /* End of for (y = 0; y < nb_lines_of_pieces; y++). */
  }      /* End of for (x = 0; x < game->nb_pieces; x++). */
  

quick_end:
  fclose (in) ;
  if (status == FALSE) {
    freeGame (game) ;
    game = NULL ;
  }
  
  return (game) ;
}
