/*
  game_save.c
  jigplay-macos
 
  Created by François Pessaux on 16/09/10.
  Copyright 2010. All rights reserved.
*/

#include <stdio.h>
#include <stdlib.h>

#include "game_save.h"
#include "alerts_n_errors.h"



static bool_t saveLayout (FILE *out, struct layout_cell *layout, int nb_cells)
{
  int i, written ;

  /* Save the list of pieces ids of each layout cell. Forget the redraw flag. */
  for (i = 0; i < nb_cells; i++) {
    struct pieces_list *list = layout[i].hosted_pieces ;
    int list_len ;

    list_len = piecesListLength (list) ;
    /* Save the number of pieces in the list of the current cell. */
    written = fwrite (&list_len, sizeof (int), 1, out) ;
    if (written != 1) return (FALSE) ;
    /* Now, save the id of each piece of the list. */
    while (list != NULL) {
      written = fwrite (&list->piece->piece_debug_id, sizeof (int), 1, out) ;
      if (written != 1) return (FALSE) ;
      list = list->next ;
    }    /* End of while (list != NULL). */
  }
  return (TRUE) ;
}



static bool_t saveSpline (FILE *out, struct spline *spline)
{
  int written ;

  /* The number of points. */
  written = fwrite (&spline->nb_points, sizeof (int), 1, out) ;
  if (written != 1) return (FALSE) ;
  /* The vertices themselves. */
  written =
    fwrite (spline->vertices, sizeof (struct point), spline->nb_points, out) ;
  if (written != spline->nb_points) return (FALSE) ;

  return (TRUE) ;
}


/** Saves all the edges present in the array passed as argument. */
static bool_t saveAllEdges (FILE *out, struct edge **edges, int nb_edges)
{
  int tmp, written, i ;

  for (i = 0; i < nb_edges; i++) {
    struct edge *edge = edges[i] ;

    /* The id. */
    written = fwrite (&edge->id, sizeof (int), 1, out) ;
    if (written != 1) return (FALSE) ;
    /* Activity. */
    written = fwrite (&edge->active, sizeof (int), 1, out) ;
    if (written != 1) return (FALSE) ;
    /* Point p0. */
    written = fwrite (&edge->p0, sizeof (struct point), 1, out) ;
    if (written != 1) return (FALSE) ;
    /* Point p0 anchor id. */
    written = fwrite (&edge->p0_anchor_id, sizeof (int), 1, out) ;
    if (written != 1) return (FALSE) ;
    /* Point p1. */
    written = fwrite (&edge->p1, sizeof (struct point), 1, out) ;
    if (written != 1) return (FALSE) ;
    /* Point p1 anchor id. */
    written = fwrite (&edge->p1_anchor_id, sizeof (int), 1, out) ;
    if (written != 1) return (FALSE) ;
    /* The spline. */
    if (saveSpline (out, edge->spline) != TRUE) return (FALSE) ;
    /* The id of the piece this edge belongs to. */
    written =
      fwrite (&edge->part_of_piece->piece_debug_id, sizeof (int), 1, out) ;
    if (written != 1) return (FALSE) ;
    
    /* The ids of our next/prev edges in the X/Y sorted lists or -1 if NULL. */
    if (edge->x_sorted_prev == NULL) tmp = -1 ;
    else tmp = edge->x_sorted_prev->id ;
    written = fwrite (&tmp, sizeof (int), 1, out) ;
    if (written != 1) return (FALSE) ;
    if (edge->x_sorted_next == NULL) tmp = -1 ;
    else tmp = edge->x_sorted_next->id ;
    written = fwrite (&tmp, sizeof (int), 1, out) ;
    if (written != 1) return (FALSE) ;
    if (edge->y_sorted_prev == NULL) tmp = -1 ;
    else tmp = edge->y_sorted_prev->id ;
    written = fwrite (&tmp, sizeof (int), 1, out) ;
    if (written != 1) return (FALSE) ;
    if (edge->y_sorted_next == NULL) tmp = -1 ;
    else tmp = edge->y_sorted_next->id ;
    written = fwrite (&tmp, sizeof (int), 1, out) ;
    if (written != 1) return (FALSE) ;
  }

  return (TRUE) ;
}



static bool_t savePieces (FILE *out, struct piece *pieces, int nb_pieces)
{
  int i, j, written ;
  struct connected_set **found_connected_sets ;
  struct edge **all_edges ;
  int nb_found_connected_sets ;

  /* Allocate an array to remind the pointers of already seen connected-sets.
     In the worst case, we will have as many connected-sets as pieces, so to
     make simpler, directly allocate an array of the maximum size. */
  found_connected_sets = malloc (nb_pieces * sizeof (struct connected_set**)) ;
  if (found_connected_sets == NULL) return (FALSE) ;
  nb_found_connected_sets = 0 ;

  /* First, identify and count the connected-sets used by the piece. We walk
     along all the pieces and remind the connected-sets appearing we didn't
     already saw. We store them in the array found_connected_sets. */
  for (i = 0; i < nb_pieces; i++) {
    int already_seen ;
    struct connected_set *current_cs = pieces[i].connected_pieces ;

    already_seen = FALSE ;
    for (j = 0; j < nb_found_connected_sets; j++) {
      if (current_cs == found_connected_sets[j]) {
        /* The current connected-set already belongs to the seen ones. */
        already_seen = TRUE ;
        break ;
      }    /* End of if (current_cs == found_connected_sets[j]). */
    }      /* End of for (j = 0; j < nb_found_connected_sets; j++). */

    /* Check if we exited without finding the current connected-set in the
       already seen ones. If so, add it to the seen ones. */
    if (already_seen != TRUE) {
      found_connected_sets[nb_found_connected_sets] = current_cs ;
      nb_found_connected_sets++ ;
    }      /* End of if (already_seen != TRUE). */
  }        /* End of for (i = 0; i < nb_pieces; i++). */

  /* Now, save the found unique connected-sets, starting by the number we
     found of. */
  written = fwrite (&nb_found_connected_sets, sizeof (int), 1, out) ;
  if (written != 1) {
    free (found_connected_sets) ;
    return (FALSE) ;
  }

  for (i = 0; i < nb_found_connected_sets; i++) {
    struct connected_set *current_cs = found_connected_sets[i] ;
    struct pieces_list *plist = current_cs->members ;
    /* Count the number of pieces in this connected-set and save it. This should
       be directly the refcount, but... it doesn't hurt. */
    int len = piecesListLength (plist) ;
    written = fwrite (&len, sizeof (int), 1, out) ;
    if (written != 1) {
      free (found_connected_sets) ;
      return (FALSE) ;
    }

    /* Now, save the refcount. */
    written = fwrite (&current_cs->refcount, sizeof (int), 1, out) ;
    if (written != 1) {
      free (found_connected_sets) ;
      return (FALSE) ;
    }

    /* Now, save the id of each piece member of this connected-set. */
    while (plist != NULL) {
      written = fwrite (&plist->piece->piece_debug_id, sizeof (int), 1, out) ;
      if (written != 1) {
        free (found_connected_sets) ;
        return (FALSE) ;
      }
      plist = plist->next ;
    }    /* End of while (plist != NULL). */
  }      /* End of for (i = 0; i < nb_found_connected_sets; i++). */

  /* Now, save all the edges present in all the pieces. We first make an array
     containing all of them. */
  all_edges = malloc (nb_pieces * 4 * sizeof (struct edges*)) ;
  if (all_edges == NULL) {
    free (found_connected_sets) ;
    return (FALSE) ;
  }
  
  /* Use j to count edges we store in the array. Walk along all the pieces,
     harvesting the pointers of their edges. */
  j = 0 ;
  for (i = 0; i < nb_pieces; i++) {
    struct piece *piece = &pieces[i] ;
    all_edges[j] = piece->free_edges[NORTH] ;
    j++ ;
    all_edges[j] = piece->free_edges[EAST] ;
    j++ ;
    all_edges[j] = piece->free_edges[SOUTH] ;
    j++ ;
    all_edges[j] = piece->free_edges[WEST] ;
    j++ ;
  }
  
  /* Now, really write the edges information. */
  written = saveAllEdges (out, all_edges, nb_pieces * 4) ;
  free (all_edges) ;
  if (! written) {
    free (found_connected_sets) ;
    return (FALSE) ;
  }
  
  /* Now, really save the pieces structures. Note that the image stuff is not
     saved, they will be rebuilt at loading-time. */
  for (i = 0; i < nb_pieces; i++) {
    struct piece *piece = &pieces[i] ;

    /* The id. */
    written = fwrite (&piece->piece_debug_id, sizeof (int), 1, out) ;
    if (written != 1) return (FALSE) ;
    /* The center. */
    written = fwrite (&piece->center, sizeof (struct point), 1, out) ;
    if (written != 1) return (FALSE) ;
    /* The bounding-box. */
    written =
      fwrite (&piece->delta_center_to_top_left, sizeof (struct point), 1, out) ;
    if (written != 1) return (FALSE) ;
    written =
      fwrite
        (&piece->delta_center_to_bottom_right, sizeof (struct point), 1, out) ;
    if (written != 1) return (FALSE) ;
    /* The edges ids. */
    for (j = 0; j < 4; j++) {
      written = fwrite (&piece->free_edges[j]->id, sizeof (int), 1, out) ;
      if (written != 1) return (FALSE) ;
    }

    /* The connected-set. We save the index of the connected set we previously
       saved that corresponds to the current piece's connected-set. Hence, we
       must recover it from the array by comparing its pointer with those stored
       in this array. */
    for (j = 0; j < nb_found_connected_sets; j++) {
      if (found_connected_sets[j] == piece->connected_pieces) {
        /* We found it, save its index in the file. */
        written = fwrite (&j, sizeof (int), 1, out) ;
        if (written != 1) {
          free (found_connected_sets) ;
          return (FALSE) ;
        }
        break ;
      }
    }    /* End of for (j = 0; j < nb_found_connected_sets; j++). */

    /* For sake of safety, ensure we really found the searched connected-set.
       This will be the case if we left the loop before it reached its end, i.e.
       before j becomes equal to nb_found_connected_sets. */
    if (j == nb_found_connected_sets) {
      /* Something went wrong, we didn't find the connected-set. */
      free (found_connected_sets) ;
      return (FALSE) ;
    }
  }

  /* Now, connected-sets and pieces with their reference on these connected-sets
     are saved. The array of found connected-sets is no more useful, free it. */
  free (found_connected_sets) ;

  return (TRUE) ;
}



bool_t saveGame (const char *file_name, struct game *game)
{
  FILE *out ;
  int tmp, written ;
  int status = TRUE ;

  out = fopen (file_name, "wb") ;
  if (out == NULL) {
    displayMessage ("Error", "Unable to open specified file.", AC_Critical) ;
    return (FALSE) ;
  }

  /* The magic number. */
  tmp = SAVEMAGICNUMBER ;
  written = fwrite (&tmp, sizeof (int), 1, out) ;
  if (written != 1) { status = FALSE ; goto quick_end ; }
  /* Miscellaneous scalars. */
  written = fwrite (&game->image_width, sizeof (int), 1, out) ;
  if (written != 1) { status = FALSE ; goto quick_end ; }
  written = fwrite (&game->image_height, sizeof (int), 1, out) ;
  if (written != 1) { status = FALSE ; goto quick_end ; }
  written =
    fwrite (&game->original_image_scale_factor, sizeof (float), 1, out) ;
  if (written != 1) { status = FALSE ; goto quick_end ; }
  written = fwrite (&game->piece_width, sizeof (int), 1, out) ;
  if (written != 1) { status = FALSE ; goto quick_end ; }
  written = fwrite (&game->piece_height, sizeof (int), 1, out) ;
  if (written != 1) { status = FALSE ; goto quick_end ; }
  written = fwrite (&game->layout_nb_columns, sizeof (int), 1, out) ;
  if (written != 1) { status = FALSE ; goto quick_end ; }
  written = fwrite (&game->layout_nb_lines, sizeof (int), 1, out) ;
  if (written != 1) { status = FALSE ; goto quick_end ; }
  written = fwrite (&game->nb_pieces, sizeof (int), 1, out) ;
  if (written != 1) { status = FALSE ; goto quick_end ; }

  /* The ids of the heads and tails of the x and y sorted lists of edges or -1
     if NULL. */
  if (game->x_sorted_edges_head == NULL) tmp = -1 ;
  else tmp = game->x_sorted_edges_head->id ;
  written = fwrite (&tmp, sizeof (int), 1, out) ;
  if (written != 1) { status = FALSE ; goto quick_end ; }
  if (game->x_sorted_edges_tail == NULL) tmp = -1 ;
  else tmp = game->x_sorted_edges_tail->id ;
  written = fwrite (&tmp, sizeof (int), 1, out) ;
  if (written != 1) { status = FALSE ; goto quick_end ; }
  if (game->y_sorted_edges_head == NULL) tmp = -1 ;
  else tmp = game->y_sorted_edges_head->id ;
  written = fwrite (&tmp, sizeof (int), 1, out) ;
  if (written != 1) { status = FALSE ; goto quick_end ; }
  if (game->y_sorted_edges_tail == NULL) tmp = -1 ;
  else tmp = game->y_sorted_edges_tail->id ;
  written = fwrite (&tmp, sizeof (int), 1, out) ;
  if (written != 1) { status = FALSE ; goto quick_end ; }
  
  /* Save pieces before, hence we will load them first and be able to create
     them to reference them by their id in the save of the layout. */
  if (! savePieces (out, game->pieces, game->nb_pieces)) {
    status = FALSE ;
    goto quick_end ;
  }

  if (! saveLayout
       (out, game->layout, game->layout_nb_columns * game->layout_nb_lines)) {
    status = FALSE ;
    goto quick_end ;
  }

  /* The image file data. */
  written = fwrite (&game->image_file_size, sizeof (int), 1, out) ;
  if (written != 1) {
    status = FALSE ;
    goto quick_end ;
  }
  
  written =
    fwrite
      (game->image_file_data, sizeof (unsigned char),
       game->image_file_size, out) ;
  if (written != game->image_file_size) status = FALSE ;
  
quick_end:
  fclose (out) ;
  return (status) ;
}
