/*
  piece.c
  jigplay-macos
 
  Created by Franois Pessaux on 29/09/10.
  Copyright 2010. All rights reserved.
*/

#include <stdlib.h>

#include "piece.h"


/** Pattern of an horizontal piece's edge. Distorsion in y with head
    pointing toward the negative coordinates. The shape is normalized
    at the origine. */
static const struct fpoint unit_horizontal_edge[NB_UNIT_EDGE_POINTS] = {
  { 0.000000, 0.000000 },  /* Start point. */
  { 0.028409, 0.000000 },
  { 0.500000, 0.000000 },
  { 0.318182, -0.136364 },
  { -0.227273 + 0.5, -0.590909 + 0.2 },
  { 1.227273 - 0.5, -0.590909 + 0.2 },
  { 0.681818, -0.136364 },
  { 0.500000, 0.000000 },
  { 0.971591, 0.000000 },
  { 1.000000, 0.000000 }    /* End point. */
};



/** [Descr] : Creates the distortion of an horizontal edge. Basically,
    we have an horizontal pattern of noise of unit size and we scale
    it to the piece width. The distortion is the moved points through
    which we will create an approximating spline in order to create
    "little ears" on our edge.
    [Params] :
     - start_x : The x coordinate where the edge starts.
     - start_y : The y coordinate where the edge starts.
     - width : The length of the edge. Hence the edge will span from
       (start_x, start_y) to (start_x + width, start_y) bounds included.
    [Returns] : A pointer representing an array of length
     NB_UNIT_EDGE_POINTS containing some struct point.
    [Rem] : Exported outside this module. */
struct point* makeHorizontalEdgeSegments (int start_x, int start_y, int width)
{
  float width_as_float = (float) width ;
  int y_flip = (rand () % 2) ? 1 : -1 ;
  struct point *result_array ;
  int i ;

  result_array = malloc (NB_UNIT_EDGE_POINTS * sizeof (struct point)) ;
  if (result_array == NULL) return (NULL) ;

  for (i = 0; i < NB_UNIT_EDGE_POINTS; i++) {
    float fx = unit_horizontal_edge[i].fx ;
    float fy = unit_horizontal_edge[i].fy ;
    
    if ((i == 4) || (i == 5)) {
      float px = /* +/- 0.2 to randomize the shape. */
        ((float) (200 - (rand () % 400))) / 1000.0 ;
      result_array[i].x = ((int) ((fx + px) * width_as_float)) + start_x ;
      result_array[i].y = y_flip * ((int) (fy * width_as_float)) + start_y ;
    }
    else {
      result_array[i].x = ((int) (fx * width_as_float)) + start_x ;
      result_array[i].y = y_flip * ((int) (fy * width_as_float)) + start_y ;
      }
  }        /* End of for (i = 0; i < NB_UNIT_EDGE_POINTS; i++). */
  
  return (result_array) ;
}



/** Pattern of an horizontal piece's edge.  Distortion in x with the
    head pointing toward the negative coordinates. The shape is
    normalised at the origine. This is simply obtained by a rotation
    of -90 (math coords system) of the above previous shape (y' = -x
    and x' = y) and a shift "up" of 1.0 to have its "lower" point
    centered on the origin. Hence, such an edge is suitable to be the
    west edge of a piece (i.e. drawn bottom up in the screen
    coordinates system). */
static const struct fpoint unit_vertical_edge[NB_UNIT_EDGE_POINTS] = {
  { 0.000000, 1.0 + 0.000000 },    /* Start point. */
  { 0.000000, 1.0 + -0.028409 },
  { 0.000000, 1.0 + -0.500000 },
  { -0.136364, 1.0 + -0.318182 },
  { -0.590909 + 0.2, 1.0 + -(-0.227273 + 0.5) },
  { -0.590909 + 0.2, 1.0 + -(1.227273 - 0.5) },
  { -0.136364, 1.0 + -0.681818 },
  { 0.000000, 1.0 + -0.500000 },
  { 0.000000, 1.0 + -0.971591 },
  { 0.000000, 1.0 + -1.000000 }     /* End point. */
};



/** [Descr] : Creates the distortion of a vertical edge. Basically,
    we have an vertical pattern of noise of unit size and we scale
    it to the piece height. The distortion is the moved points through
    which we will create an approximating spline in order to create
    "little ears" on our edge.
    [Params] :
     - start_x : The x coordinate where the edge starts.
     - start_y : The y coordinate where the edge starts.
     - height : The height of the edge. Hence the edge will span from
       (start_x, start_y) to (start_x, start_y + height) bounds included.
    [Rem] : Exported outside this module. */
struct point* makeVerticalEdgeSegments (int start_x, int start_y, int height)
{
  float height_as_float = (float) height ;
  int x_flip = (rand () % 2) ? 1 : -1 ;
  struct point *result_array ;
  int i ;

  result_array = malloc (NB_UNIT_EDGE_POINTS * sizeof (struct point)) ;
  if (result_array == NULL) return (NULL) ;

  for (i = 0; i < NB_UNIT_EDGE_POINTS; i++) {
    float fx = unit_vertical_edge[i].fx ;
    float fy = unit_vertical_edge[i].fy ;

    if ((i == 4) || (i == 5)) {
      float py = /* +/- 0.2 to randomize the shape. */
        ((float) (200 - (rand () % 400))) / 1000.0 ;
      result_array[i].x = (x_flip * ((int) (fx * height_as_float))) + start_x ;
      result_array[i].y = ((int) ((fy + py) * height_as_float)) + start_y ;
    }
    else {
      result_array[i].x = x_flip * ((int) (fx * height_as_float)) + start_x ;
      result_array[i].y = ((int) (fy * height_as_float)) + start_y ;
    }
  }
  
  return (result_array) ;
}



/** Releases memory occupied by a list of pieces, but not the pieces
    themselves. */
void freePiecesList (struct pieces_list *lst)
{
  struct pieces_list *tmp ;
  
  while (lst != NULL) {
    tmp = lst->next ;
    free (lst) ;
    lst = tmp ;
  }
}



struct pieces_list* copyPiecesList  (struct pieces_list *lst)
{
  struct pieces_list *copy_head, *tmp, *prev ;
  
  copy_head = NULL ;
  prev = NULL ;
  
  while (lst != NULL) {
    tmp = malloc (sizeof (struct pieces_list)) ;
    tmp->piece = lst->piece ;
    tmp->next = NULL ;
    if (copy_head == NULL) copy_head = tmp ;
    if (prev != NULL) prev->next = tmp ;
    prev = tmp ;
    lst = lst->next ;
  }
  
  return (copy_head) ;
}



/** Releases memory occupied by a list of edges, but not the edges
    themselves. */
void freeEdgesList (struct edges_list *lst)
{
  struct edges_list *tmp ;
  
  while (lst != NULL) {
    tmp = lst->next ;
    free (lst) ;
    lst = tmp ;
  }
}


int piecesListLength (struct pieces_list *list)
{
  int len = 0 ;
  while (list != NULL) {
    len++ ;
    list = list->next ;
  }
  
  return (len) ;
}
