/*
  piece.h
  jigplay-macos
 
  Created by François Pessaux on 29/09/10.
  Copyright 2010. All rights reserved.
*/

#ifndef __PIECE_H__
#define __PIECE_H__

#ifdef __JIGPLAY_MAC_OS__
#include <Carbon/Carbon.h>
#else
#ifdef __JIGPLAY_IOS__
#include <CoreGraphics/CGImage.h>
#endif
#endif

#include "maths2d.h"
#include "spline.h"
#include "image_with_mask.h"


/** Represents an edge of a piece of the jigsaw. Each piece own 4
    edges.
    Anchor points ids layout:
    0- 1- 2- 3- 4
    |  |  |  |  |
    5- 6- 7- 8-10
    |  |  |  |  |
    11-12-13-14-15 */
struct edge {
  int id ;
  /** Boolean flag telling if the edge is still important and must be
      taken into account during computations. */
  int active ;
  /** Coordinate in the screen space of the first point of the edge. */
  struct point p0 ;
  /** Unique id representing the first point of the edge. This point
      is an anchor point and must match another anchor point of another
      piece in order to have pieces coalescing. */
  int p0_anchor_id ;
  /* Coordinate in the screen space of the second point of the edge. */
  struct point p1 ;
  /** Unique id representing the second point of the edge. This point
      is an anchor point and must match another anchor point of another
      piece in order to have pieces coalescing. */
  int p1_anchor_id ;
  /** The spline representing the shape of the piece. */
  struct spline *spline ;
  /** Pointer to the piece this edge belongs to. */
  struct piece *part_of_piece ;
  /** Pointer representing this edge's belonging to a double linked
      list where edges are sorted bu increasing minimal x
      coordinate. This pointer points to the next edge, having a
      minimal x coordinate greated or equal to the current edge's
      one. This list is used to find faster which edges are close to
      an edge by walking forward and backward in this list from the
      currently selected piece. */
  struct edge *x_sorted_prev ;
  /** Pointer representing this edge's belonging to a double linked
      list where edges are sorted by increasing minimal x
      coordinate. This pointer points to the previous edge, having a
      minimal x coordinate lower or equal to the current edge's
      one. This list is used to find faster which edges are close to
      an edge by walking forward and backward in this list from the
      currently selected piece. */
  struct edge *x_sorted_next ;
  /** Pointer representing this edge's belonging to a double linked
      list where edges are sorted by increasing minimal y
      coordinate. This pointer points to the next edge, having a
      minimal y coordinate greated or equal to the current edge's
      one. This list is used to find faster which edges are close to
      an edge by walking forward and backward in this list from the
      currently selected piece. */
  struct edge *y_sorted_prev ;
  /** Pointer representing this edge's belonging to a double linked
      list where edges are sorted bu increasing minimal y
      coordinate. This pointer points to the previous edge, having a
      minimal y coordinate lower or equal to the current edge's
      one. This list is used to find faster which edges are close to
      an edge by walking forward and backward in this list from the
      currently selected piece. */
  struct edge *y_sorted_next ;
};



#define NORTH (0)
#define EAST (1)
#define SOUTH (2)
#define WEST (3)

/** Represents a 4-sided piece of the jigsaw. */
struct piece {
  int piece_debug_id ;
  /** Position of the center of the piece in a graphic coordinates
      system, not in a mathematical one. Used to faster determine
      which piece is under the mouse pointer. */
  struct point center ;
  /** Delta in x and y of the top left point of the piece's bounding
      box regarding to the center of the piece. */
  struct point delta_center_to_top_left ;
  /** Delta in x and y of the bottom right point of the piece's bounding
      box regarding to the center of the piece. */
  struct point delta_center_to_bottom_right ;
  /** Initial order N-E-S-O (indices 0-1-2-3) form a polygon. Once an
      edge matches another, hence 2 pieces coalesce, these 2 edges
      are destroyed from the pieces and removed from the double-linked
      lists to save computation. The edges array always contains 4 edges. */
  struct edge *free_edges[4] ;
  /** Reference on the connected-set this piece belongs to if this piece
      is connected to other pieces. At least contains the piece itself. */
  struct connected_set *connected_pieces ;
  struct image_with_mask *piece_image ;
};



struct edges_list {
  struct edge *edge ;
  struct edges_list *next ;
};



struct pieces_list {
  struct piece *piece ;
  struct pieces_list *next ;
};



struct connected_set {
  /** The list of all the pieces connected together. */
  struct pieces_list *members ;
  /* Count the number of pieces structures pointing (sharing) this connected-set
     structure. This is used to prevent disallocation while some piece
     structures still point on the connected-set and also allows to free it
     only once.
     Moreover, this allows to know when the game ends: there must remain only
     one connected-set, hence the refcount must be equal to the number of pieces
     of the game. */
  int refcount ;
};

/** Size of the array returned by make_horizontal_edge_segments and
    make_vertical_edge_segments. */
#define NB_UNIT_EDGE_POINTS 10
struct point* makeHorizontalEdgeSegments (int start_x, int start_y, int width) ;
struct point* makeVerticalEdgeSegments (int start_x, int start_y, int height) ;
void freePiecesList (struct pieces_list *lst) ;
void freeEdgesList (struct edges_list *lst) ;
struct pieces_list* copyPiecesList  (struct pieces_list *lst) ;
void free_piece (struct piece *piece) ;
int piecesListLength (struct pieces_list *list) ;

#endif
