/*
  spline.c
  jigplay-macos
 
  Created by Franois Pessaux on 29/09/10.
  Copyright 2010. All rights reserved.
*/

#include <stdlib.h>

#include "misc.h"
#include "spline.h"


/** Helper to compute the middle point of a segment. */
void middle (struct point *p0, struct point *p1, struct point *result)
{
  result->x = (p0->x + p1->x) >> 1 ;
  result->y = (p0->y + p1->y) >> 1 ;
}



/** [Descr] : Fonction qui approxime une courbe de Bzier par des
    segments de droite.
    Retourne le tableau de points reprsentant la courbe ainsi que les
    2 points extrmites de la bounding box englobant la courbe.
    [Rem] : From Michel Quercia's library "graphx", in the graph.ml
    file. */
struct spline* buildSpline (int rec_depth, struct point *vertices,
                            int nb_vertices, struct point *top_left,
                            struct point *bottom_right)
{
  int lr, i, l, j ;
  struct point *spline_vertices ;
  struct spline *spline ;

  /* Control the number of points and allocate the array of result points. */
  if ((nb_vertices % 3) != 1) return (NULL) ;

  lr = (nb_vertices - 1) << rec_depth ;
  spline_vertices = malloc ((lr + 1) * sizeof (struct point)) ;
  if (spline_vertices == NULL) return (NULL) ;
  spline = malloc (sizeof (struct spline)) ;
  if (spline == NULL) {
    free (spline_vertices) ;
    return (NULL) ;
  }
  spline->nb_points = lr + 1 ;
  spline->vertices = spline_vertices ;

  /* Distribue les points initiaux : les coordonnes sont multiplies par */
  /* 2^16 pour amliorer la prcision dans le calcul des middles.         */
  for (i = 0; i < nb_vertices; i++) {
    spline_vertices[i].x = vertices[i].x << 16 ;
    spline_vertices[i].y = vertices[i].y << 16 ;
  }

  /* Divise les arcs [rec_depth] fois de suite. */
  l = nb_vertices - 1 ;
  while (l < lr) {
    i = l ;
    j = l << 1 ;
    l = j ;
    while (i > 0) {
      struct point ab, bc, cd, abc, bcd, abcd ;

      struct point *a = &spline_vertices[i] ;
      struct point *b = &spline_vertices[i - 1] ;
      struct point *c = &spline_vertices[i - 2] ;
      struct point *d = &spline_vertices[i - 3] ;

      middle (a, b, &ab) ;
      middle (b, c, &bc) ;
      middle (c, d, &cd) ;
      middle (&ab, &bc, &abc) ;
      middle (&bc, &cd, &bcd) ;
      middle (&abc, &bcd, &abcd) ;

      spline_vertices[j].x = a->x ;
      spline_vertices[j].y = a->y ;
      spline_vertices[j - 1].x = ab.x ;
      spline_vertices[j - 1].y = ab.y ;
      spline_vertices[j - 2].x = abc.x ;
      spline_vertices[j - 2].y = abc.y ;
      spline_vertices[j - 3].x = abcd.x ;
      spline_vertices[j - 3].y = abcd.y ;
      spline_vertices[j - 4].x = bcd.x ;
      spline_vertices[j - 4].y = bcd.y ;
      spline_vertices[j - 5].x = cd.x ;
      spline_vertices[j - 5].y = cd.y ;
      
      i -= 3 ;
      j -= 6 ;
      /* Rmq : inutile de recopier le 1er point, [a], il sera mis en place
         lors de l'itration suivante ou il est dj en place si c'est
         spline_vertices.(0) */
    }    /* End of while (i > 0). */
  }      /* End of while (l < lr). */

  /* Divise les points rsultant par 2^16 pour revenir  l'chelle initiale. */
#define C (1 << 15)
  /* Initialize the bounding box with the first vertex of the polyline used to
     draw the spline. This point will anyway belong to the spline, hence will
     anyway be included in the bounding box. */
  top_left->x = vertices[0].x ;
  top_left->y = vertices[0].y ;
  bottom_right->x = vertices[0].x ;
  bottom_right->y = vertices[0].y ;

  for (i = 0 ; i <= lr; i++) {
    int point_x = (spline_vertices[i].x + C) >> 16 ;
    int point_y = (spline_vertices[i].y + C) >> 16 ;
    
    spline_vertices[i].x = point_x ;
    spline_vertices[i].y = point_y ;
    /* Update the bounding box. */
    top_left->x = min (top_left->x, point_x) ;
    top_left->y = min (top_left->y, point_y) ;
    bottom_right->x = max (bottom_right->x, point_x) ;
    bottom_right->y = max (bottom_right->y, point_y) ;
  }
  
  return (spline) ;
}



/** shift_up: used to tell if copy is used to make the spline of a
    shouth upper piece, hence points must have a y coordinate 1
    lower of if the copy is used to make the spline of a west
    lefter piece and hence points must have a x coordinate 1
    lower. If shift_up is true then first case, else second case. */
struct spline* build_opposite_spline (int shift_up, struct spline *spline)
{
  int i, x_shift, y_shift ;
  struct spline *new_spline ;
  struct point *new_points ;
  int nb_points = spline->nb_points ;
  struct point *points = spline->vertices ;

  new_points = malloc (nb_points * sizeof (struct point)) ;
  if (new_points == NULL) return (NULL) ;
  new_spline = malloc (sizeof (struct spline)) ;
  if (new_spline == NULL) {
    free (new_points) ;
    return (NULL) ;
  }
  new_spline->nb_points = nb_points ;
  new_spline->vertices = new_points ;

  if (shift_up) {
    x_shift = 0 ;
    y_shift = -1 ;
  }
  else {
    x_shift = -1 ;
    y_shift = 0 ;
  }

  /* Simply reverse the order of the points of the initial spline. */
  for (i = 0; i < nb_points; i++) {
    new_points[i].x = points[nb_points - 1 - i].x + x_shift ;
    new_points[i].y = points[nb_points - 1 - i].y + y_shift ;
  }

  return (new_spline) ;
}
