/*
  texturemap.c
  jigplay-macos
 
  Created by François Pessaux on 29/09/10.
  Copyright 2010. All rights reserved.
*/

#include <stdlib.h>

#include "misc.h"
#include "texturemap.h"
#include "image_with_mask.h"



/** Function used to pass to [qsort] in order to sort some floats
    in increasing order.
    Returns:
    0 if flt1's criterion is equal to flt2's critrion
    < 0 if flt1's criterion is lower to flt2's criterion
    > 0 if flt1's criterion is greater to flt2's critrion. */
static int intersec_increasing_compare_fct (const float *flt1, const float *flt2)
{
  return (*flt1 - *flt2) ;
}



/* just to shorten usage of the [qsort] function. */
typedef int (*ComparisonFct)(const void*, const void*) ;



static bool_t fillMask (unsigned char *dest_image_bits,
                        int dest_width, int dest_height, int dest_width_in_bytes,
                        struct segment *segments, int nbseg)
{
  int i, j, k, yh, left, right, cpt, aux ;
  int *indir ;
  float *invp ;
  float *intersec ;

  indir = malloc (nbseg * sizeof (int)) ;
  if (indir == NULL) return (FALSE) ;
  invp = malloc (nbseg * sizeof (float)) ;
  if (invp == NULL) {
    free (indir) ;
    return (FALSE) ;
  }
  intersec = malloc (nbseg * sizeof (float)) ;
  if (intersec == NULL) {
    free (indir) ;
    free (invp) ;
    return (FALSE) ;
  }

  /* Indirection table allowing to sort edges by increasing y initially
     initialized with identity. */
  for (i = 0; i < nbseg; i++) indir[i] = i ;

  /* Make sure that for each edge, the point with the minimal y is the
     first one (the start point). */
  for (i = 0; i <nbseg; i++) {
    if (segments[i].p0.y > segments[i].p1.y) {
      int x_max = segments[i].p0.x ;
      int y_max = segments[i].p0.y ;
      int x_min = segments[i].p1.x ;
      int y_min = segments[i].p1.y ;
      segments[i].p0.x = x_min ;
      segments[i].p0.y = y_min ;
      segments[i].p1.x = x_max ;
      segments[i].p1.y = y_max ;
    }
  }
  /* Remove horizontal edges and shift remaining edges to the left
     in the indirection array (do not modify the edges themselves). */
  i = 0 ;
  while (i <= (nbseg - 1)) {
    if (segments[indir[i]].p0.y == segments[indir[i]].p1.y) {
      for (j = i + 1; j < nbseg; j++) {
        indir[j - 1] = indir[j] ;
      }
      nbseg-- ;
    }
    else i++ ;
  }
  /* Sort edges in the indirection table so that y are in increasing
     order. */
  for (k = nbseg - 1; k >= 1; k--) {
    int indmax = 0 ;
    for (j = 1; j <= k; j++) {
      if (segments[indir[indmax]].p0.y < segments[indir[j]].p0.y) {
        indmax = j ;
      }
      aux = indir[indmax] ;
      indir[indmax] = indir[k] ;
      indir[k] = aux ;
    }    /* End of for (j = 1; j <= k; j++). */
  }      /* End of for (k = nbseg - 1; k >= 1; k--). */
  /* */
  for (i = 0; i < nbseg; i++) {
    invp[indir[i]] =
      ((float) (segments[indir[i]].p1.x - segments[indir[i]].p0.x))
      /
      ((float) (segments[indir[i]].p1.y - segments[indir[i]].p0.y)) ;
  }
  /* */
  yh = segments[indir[0]].p0.y ;
  right = 0 ;
  left = -1 ;
  while (left < (nbseg - 1)) {
    while ((segments[indir[right]].p0.y <= yh) &&
           (right <= (nbseg - 1))) {
      right = right + 1 ;
    }
    for (k = left + 1; k <= (right - 1); k++) {
      if (segments[indir[k]].p1.y <= yh) {
        left = left + 1 ;
        aux = indir[left] ;
        indir[left] = indir[k] ;
        indir[k] = aux ;
      }
    }
    /* */
    cpt = 0 ;
    for (k = left + 1; k <= (right - 1); k++) {
      intersec[cpt] =
        ((float) segments[indir[k]].p0.x) +
        ((float) (yh - segments[indir[k]].p0.y)) *
        invp[indir[k]] ;
      cpt++ ;
    }
    if (cpt > 0) {
      int line, x ;

      /* Sort the array on the cpt first elements, in increasing order. */
      qsort
        (intersec, cpt, sizeof (float),
         (ComparisonFct) intersec_increasing_compare_fct) ;

      line = 0 ;
      while (line <= (cpt - 1)) {
        int x_start = (int) intersec[line] ;
        int x_stop = (int) intersec[line + 1] ;

        /* Draw an horizontal line. */
        for (x = x_start; x <= x_stop; x++) {
          /* Plot at [x][yh] from source
             [x + x_source_offset][yh_in_source]. */
          int bit_index ;
          unsigned char bit_mask ;
          if ((yh >= 0) && (yh < dest_height) &&
              (x >= 0) && (x < dest_width)) {
            bit_index = (yh * dest_width_in_bytes) + (x / 8) ;
            bit_mask = ~(((unsigned char) 1) << (7 - (x % 8))) ;
            dest_image_bits[bit_index] &= bit_mask ;
          }
        }
        line += 2 ;
      }
    }
    yh = yh++ ; /* Next line. */
  }      /* End of while (left < (nbseg - 1)). */

  free (indir) ;
  free (invp) ;
  free (intersec) ;
  return (TRUE) ;
}



/** Function used to create the piece texture with transparent areas. It maps
    the related original image part of the picture to the piece and makes
    transparent area outside the interior of the polygon described by the
    splines of the piece's edges. */
bool_t initializePieceTexture (CGImageRef original_image,
                               struct piece *piece,
                               struct point *piece_initial_center)
{
  int i, j, count ;
  int x_shift, y_shift ;
  int dest_img_width, dest_img_height ;
  int x_source_offset, y_source_offset ;
  int mask_size, bytes_per_row ;
  unsigned char *mask_bits ;
  struct segment *segments ;
  CGRect sub_image_rect ;
  CGImageRef non_masked_image, mask_image, masked_image ;
  CGDataProviderRef provider ;
  struct image_with_mask *iwm ;
  
  /* Count the total number of edges in all the splines of the
     piece. */
  int nb_edges = 0 ;
  for (i = 0; i < 4; i++) {
    /* Attention, the last point doesn't count for an edge. Problem
      of posts and intervals... */
    nb_edges += (piece->free_edges[i]->spline->nb_points - 1) ;
  }

  /* Create the array of segments to store the polygon description
     issued from the catenation of all the edges of the piece. We
     initialize the points coordinates with a default value and we
     will update it later. */
  segments = malloc (nb_edges * sizeof (struct segment)) ;
  if (segments == NULL) return (FALSE) ;

  /* Compute the delta in x and y to substract to each segment in
     order to simulate the fact that the piece is at (0, 0). Hence,
     filling routine will fill in the piece's image which is a simple
     matrix starting at (0, 0) and ending at (piece_width, piece_height)
     with boundaries included. */
  x_shift = piece_initial_center->x + piece->delta_center_to_top_left.x ;
  y_shift = piece_initial_center->y + piece->delta_center_to_top_left.y ;
  /* Update the segments coordinates from the splines points
     coordinates. */
  count = 0 ;
  for (j = 0;  j < 4; j++) {
    struct edge *edge = piece->free_edges[j] ;
    for (i = 0; i < edge->spline->nb_points - 1; i++) {
      segments[count].p0.x = edge->spline->vertices[i].x - x_shift ;
      segments[count].p0.y = edge->spline->vertices[i].y - y_shift ;
      segments[count].p1.x = edge->spline->vertices[i + 1].x - x_shift ;
      segments[count].p1.y = edge->spline->vertices[i + 1].y - y_shift ;
      count++ ;
    }
  }     /* End of for (j = 0;  j < 4; j++). */

  /* Compute the with and height of the area to map, i.e. the piece's extend. */
  dest_img_height =
    piece->delta_center_to_bottom_right.y -
    piece->delta_center_to_top_left.y ;
  dest_img_width =
    piece->delta_center_to_bottom_right.x -
    piece->delta_center_to_top_left.x ;
  
  /* We need the x and y offsets in the source image where to find colors to put
     in the destination image. In effect, the destination image space spans from
     (0, 0) to (piece_width, piece_height) but its points' colors must be taken
     in the source image at the place the piece lays. */
  x_source_offset =
    piece_initial_center->x + piece->delta_center_to_top_left.x ;
  y_source_offset =
    piece_initial_center->y + piece->delta_center_to_top_left.y ;
    
  /* Extract the non-masked image from the global picture. */
  sub_image_rect =
    CGRectMake
      (x_source_offset, y_source_offset, dest_img_width, dest_img_height) ;
  non_masked_image =
    CGImageCreateWithImageInRect (original_image, sub_image_rect) ;

  /* Allocate bytes to store the mask's bits. */
  bytes_per_row = dest_img_width / 8 ;
  if ((dest_img_width % 8) != 0) bytes_per_row++ ;
  mask_size = bytes_per_row * dest_img_height ;
  mask_bits = malloc (mask_size * sizeof (unsigned char)) ;
  if (mask_bits == NULL) {
    CGImageRelease (non_masked_image) ;
    return (FALSE) ;
  }
  
  /* Pre-clear the mask's bits to set all to transparent. */
  for (i = 0; i < mask_size; i++) mask_bits[i] = 0xFF ;

  /* Fill the mask following the splines shape and release the segments since
     they are no more useful. */
  fillMask
    (mask_bits, dest_img_width, dest_img_height, bytes_per_row, segments,
     nb_edges) ;
  free (segments) ;
  
  /* Create the mask image. */
  provider =
    CGDataProviderCreateWithData
      (NULL, mask_bits, mask_size * sizeof (unsigned char), NULL) ;
  if (provider == NULL) {
    CGImageRelease (non_masked_image) ;
    free (mask_bits) ;
    return (FALSE) ;
  }
  mask_image =
    CGImageMaskCreate
      (dest_img_width, dest_img_height, 1, 1, bytes_per_row, provider,
       NULL, 0) ;
  CGDataProviderRelease (provider) ;
  
  /* Create the masked image. */
  masked_image = CGImageCreateWithMask (non_masked_image, mask_image) ;
    
  iwm = malloc (sizeof (struct image_with_mask)) ;
  if (iwm == NULL) {
    CGImageRelease (non_masked_image) ;
    CGImageRelease (mask_image) ;
    CGImageRelease (masked_image) ;
    free (mask_bits) ;
    return (FALSE) ;
  }
  iwm->mask_bits = mask_bits ;
  iwm->mask_image = mask_image ;
  iwm->masked_image = masked_image ;
  
  /* Finally, set the image information in the piece's structure. */
  piece->piece_image = iwm ;
  
  /* Non masked image is not more useful. */
  CGImageRelease (non_masked_image) ;
  
  return (TRUE) ;
}



void freeImageWithMask (struct image_with_mask *iwm)
{
  if (iwm != NULL) {
    if (iwm->masked_image != NULL) CGImageRelease (iwm->masked_image) ;
    if (iwm->mask_image != NULL) CGImageRelease (iwm->mask_image) ;
    if (iwm->mask_bits != NULL) free (iwm->mask_bits) ;
    free (iwm) ;
  }
}
