/*
  ColorPickerView.m
 
  Created by François Pessaux on 16/10/10.
  Copyright 2010. All rights reserved.
*/

#import "ColorPickerView.h"

#include "misc.h"

@implementation ColorPickerView


#pragma mark -
#pragma mark Initialization of offscreen data

- (bool_t) initOffScreenImageAndData
{
  CGImageRef image = colorsImageView.image.CGImage ;  
  colors_image_width = (int) CGImageGetWidth (image) ;
	int height = (int) CGImageGetHeight (image) ;
  color_space = CGColorSpaceCreateDeviceRGB () ;
  
  int bitmapBytesPerRow = (colors_image_width * 4) ;
	int bitmapByteCount = (bitmapBytesPerRow * height) ;

  /* Allocate memory to record the to the image data. */
  colorsRawData = malloc (bitmapByteCount * sizeof (unsigned char)) ;
  if (colorsRawData == NULL) {
    CGColorSpaceRelease (color_space) ;
    color_space = NULL ;
    return (FALSE) ;
  }
  
  /* Create off screen bitmap context to draw the image into. */
	offscreen_context =
    CGBitmapContextCreate
      (colorsRawData, colors_image_width, height, 8, bitmapBytesPerRow,
       color_space, kCGImageAlphaPremultipliedFirst) ;
	if (offscreen_context == NULL) {
    free (colorsRawData) ;
    colorsRawData = NULL ;
    CGColorSpaceRelease (color_space) ;
    color_space = NULL ;
    return (FALSE) ;
  }

  /* Draw the image to the bitmap context. Once we draw, the memory allocated
     for the context for rendering will then contain the raw image data in the
     specified color space. */  
  CGContextDrawImage
    (offscreen_context, CGRectMake (0, 0, colors_image_width, height), image) ;

  colorsImageTopLeftPoint = colorsImageView.frame.origin ;
  colorsImageBottomRightPoint =
    CGPointMake
      (colorsImageTopLeftPoint.x + colorsImageView.frame.size.width,
       colorsImageTopLeftPoint.y + colorsImageView.frame.size.height) ;

  return (TRUE) ;
}



- (void) awakeFromNib
{
  [self initOffScreenImageAndData] ;
}



#pragma mark -
#pragma mark Free Offscreen Data and Disallocation

- (void) freeOffScreenImageAndData
{
  if (colorsRawData != NULL) {
    free (colorsRawData) ;
    colorsRawData = NULL ;
  }
  if (offscreen_context != NULL) {
    CGContextRelease (offscreen_context) ;
    offscreen_context = NULL ;
  }
  if (color_space != NULL) {
    CGColorSpaceRelease (color_space) ;
    color_space = NULL ;
  }
}



- (void) dealloc
{
  [self freeOffScreenImageAndData] ;
  [super dealloc] ;
}



/** Mostly inspired from http://www.markj.net/iphone-uiimage-pixel-color/ */
- (UIColor*) getPixelColorAtLocation: (CGPoint) point
{
  UIColor *color = nil ;
  
	if (colorsRawData != NULL) {
		/* offset locates the pixel in the data from x, y. 4 for 4 bytes of data per
       pixel, width is the width of one row of data. */
		int offset =
      4 * ((colors_image_width * ((int) (point.y))) + ((int) (point.x))) ;
		// int alpha =  colorsRawData[offset] ;
		unsigned char red = colorsRawData[offset + 1] ;
		unsigned char green = colorsRawData[offset + 2] ;
		unsigned char blue = colorsRawData[offset + 3] ;
		color =
      [UIColor colorWithRed: (((float) red) / 255.0f)
                      green: (((float) green) / 255.0f)
                       blue: (((float) blue) / 255.0f) alpha: 1.0f] ;
	}
  
	return (color) ;
}


/** Since the job is the same in case of touch started or moved, this function
    factorizes the code. It simply ensure that the touch was in the colors
    image area and update the color of the label providing visual feedback. */
- (void) touchesBeganOrMoved: (NSSet*) touches
{
  UITouch *touch = [touches anyObject] ;
  CGPoint loc = [touch locationInView: self] ;
  int x = (int) loc.x ;
  int y = (int) loc.y ;
  
  if ((x >= colorsImageTopLeftPoint.x) && (x < colorsImageBottomRightPoint.x) &&
      (y >= colorsImageTopLeftPoint.y) && (y < colorsImageBottomRightPoint.y)) {
    /* Correct the location to make it relative to the colors image dimensions.
       Hence, simply substract the colors ImageView position. */
    loc.x = loc.x - (int) colorsImageView.frame.origin.x ;
    loc.y = loc.y - (int) colorsImageView.frame.origin.y ;
    UIColor *color = [self getPixelColorAtLocation: loc] ;
    [colorSampleLabel setBackgroundColor: color] ;
    [okButton setEnabled: TRUE] ;
  }  
}

  

- (void) touchesMoved: (NSSet*) touches withEvent: (UIEvent*) event
{
  [self touchesBeganOrMoved: touches] ;
}



- (void) touchesBegan: (NSSet*) touches withEvent: (UIEvent*) event
{
  [self touchesBeganOrMoved: touches] ;
}



/** Callback used when the cancel button is clicked. Since the color picker can
    only be invoked when in game mode, we can simply restore the play view. */
- (IBAction) cancelButtonPressed:(id) sender
{
  [viewController showFramePlayAreaView] ;
}



/** Callback used when the OK button is clicked. Since the color picker can
    only be invoked when in game mode, we can simply restore the play view. */
- (IBAction) okButtonPressed:(id) sender
{
  UIColor *color =  colorSampleLabel.backgroundColor ;
  
  if (color != NULL) {
    [[viewController.framePlayAreaView getPlayAreaView]
     setBackgroundColor: color] ;
  }
  [viewController showFramePlayAreaView] ;
}

@end
