/*
  PlayAreaView.m

  Created by François Pessaux on 09/10/10.
  Copyright 2010. All rights reserved.
*/

#import "alerts_n_errors.h"
#import "PlayAreaView.h"

#include <QuartzCore/CALayer.h>
#include <QuartzCore/CATransaction.h>

#include "spline.h"
#include "piece.h"
#include "game_struct.h"
#include "game_alloc.h"
#include "game_free.h"
#include "game_manips.h"
#include "game_save.h"
#include "game_load.h"

@implementation PlayAreaView



/** Initializes the GUI part of the game, i.e. graphical elements that show
    pieces on the view. This function must be called once new_game was called
    and not before otherwise the game structure won't be initialized.
    It records all the layers allocated for pieces. Since pieces array is
    indexed by pieces ids, so will be the layers array. */
- (void) initGamePiecesLayers: (struct game*) game
{
  int i ;
  
  if (game == NULL) return ;
  piecesLayers = malloc (game->nb_pieces * sizeof (CALayer*)) ;
  
  for (i = 0; i < game->nb_pieces; i++) {
    struct piece *curr_piece = &game->pieces[i] ;
    CGRect pieceLayerframe ;
    CALayer *pieceLayer = [CALayer layer] ;
    
    /* Set the position and dimension of the layer. */
    pieceLayerframe.origin.x =
      curr_piece->center.x + curr_piece->delta_center_to_top_left.x ;
    pieceLayerframe.origin.y =
      curr_piece->center.y + curr_piece->delta_center_to_top_left.y;
    pieceLayerframe.size.width =
      curr_piece->delta_center_to_bottom_right.x -
      curr_piece->delta_center_to_top_left.x ;
    pieceLayerframe.size.height =
      curr_piece->delta_center_to_bottom_right.y -
      curr_piece->delta_center_to_top_left.y ;
    pieceLayer.frame = pieceLayerframe ;
    
    /* Don't flip the layer's coordinates system. On iOS, coordinate system is
       already with origin at the top-left corner. */
    pieceLayer.contents = (id) curr_piece->piece_image->masked_image ;
    
    [playAreaRootLayer addSublayer: pieceLayer] ;
    
    /* Remind this layer to be able to release it later and to be able to
       access it via its piece's id. */
    piecesLayers[curr_piece->piece_debug_id] = pieceLayer ;
  }
}



/** Release memory and graphical objects that represent pieces on the view. This
    function must be called before the game structure is freed. */
- (void) freeGamePiecesLayers:  (struct game*) game
{
  int i ;
  
  if (piecesLayers != NULL) {
    for (i = 0; i < game->nb_pieces; i++) {
      piecesLayers[i].contents = nil ;
      /* Remove the layer from its parent. */
      [piecesLayers[i] removeFromSuperlayer] ;
    }
    
    free (piecesLayers) ;
    piecesLayers = NULL ;
  }
}



/** Refreshes the display by redrawing only areas corresponding to all the
    pieces belonging to the same connected-set than the one passed as
    argument. */
- (void) partialUpdateDisplay: (struct game*) game
                    fromPiece: (struct piece*) from_piece
{
  int x, y ;
  
  /* First, determine which part of the display must be cleared. This
     corresponds to the union of the bounding boxes of all the pieces belonging to
     the connected-set of the piece passed as argument. While determining this
     part of the display, we also return the list of layout cells that must be
     refreshed. These cells are all those on which a piece to refresh spans.
     Todo BUG !!! We do not actually also refresh things at the place where the
     piece and its connected-set members were before arriving at their new
     position !!! */
  struct pieces_list *lst = from_piece->connected_pieces->members ;
  while (lst != NULL) {
    struct piece *piece = lst ->piece ;
    int x0 = piece->center.x + piece->delta_center_to_top_left.x ;
    int y0 = piece->center.y + piece->delta_center_to_top_left.y ;
    int x1 = piece->center.x + piece->delta_center_to_bottom_right.x ;
    int y1 = piece->center.y + piece->delta_center_to_bottom_right.y ;
    int x0_cell = x0 / game->piece_width ;
    int y0_cell = y0 / game->piece_height ;
    int x1_cell = x1 / game->piece_width ;
    int y1_cell = y1 / game->piece_height ;
    
    for (x = x0_cell; x <= x1_cell; x++) {
      for (y = y0_cell; y <= y1_cell; y++) {
        if ((x >= 0) && (x < game->layout_nb_columns) &&
            (y >= 0) && (y < game->layout_nb_lines)) {
          /* Better than creating a list of cells coordinates to redraw,
             directly mark the cells. */
          game->layout[(y * game->layout_nb_columns) + x].to_redraw = TRUE ;
        }
      }
    }
    lst = lst->next ;
  }
  
  
  /* We don't want animation. */
  [CATransaction flush] ;
  [CATransaction begin] ;
  [CATransaction setValue: (id) kCFBooleanTrue
                   forKey: kCATransactionDisableActions] ;  
  /* At this point, the display is cleared where required and the list
     of cells to redraw in the layout is known. Just walk along this
     list an redraw all the pieces of the layout in the found cells. */
  for (x = 0; x < game->layout_nb_columns; x++) {
    for (y = 0; y < game->layout_nb_lines; y++) {
      int index = (y * game->layout_nb_columns) + x ;
      
      /* If the cell is marked to be redrawn... */
      if (game->layout[index].to_redraw) {
        /* ... redraw all the pieces inside this cell... */
        lst = game->layout[index].hosted_pieces ;
        while (lst != NULL) {
          struct piece *curr_piece = lst->piece ;
          CGRect pieceLayerframe ;
          
          pieceLayerframe.origin.x =
          curr_piece->center.x + curr_piece->delta_center_to_top_left.x ;
          pieceLayerframe.origin.y =
          curr_piece->center.y + curr_piece->delta_center_to_top_left.y;
          pieceLayerframe.size.width =
          curr_piece->delta_center_to_bottom_right.x -
          curr_piece->delta_center_to_top_left.x ;
          pieceLayerframe.size.height =
          curr_piece->delta_center_to_bottom_right.y -
          curr_piece->delta_center_to_top_left.y ;
          piecesLayers[lst->piece->piece_debug_id].frame = pieceLayerframe ;
          
          lst = lst->next ;
        }
        /* ... and mark the cell as not anymore needing redraw. */
        game->layout[index].to_redraw = FALSE ;
      }
    }
  }
  [CATransaction commit] ;
}



- (void) endCurrentGameAndCleanup
{
  /* If a game is already started, kill it and release its memory. */
  if (current_game != NULL) {
    [self freeGamePiecesLayers: current_game] ;
    freeGame (current_game) ;
    current_game = NULL ;
  }
  
  /* Attention: since on iOS we do not allocate the playAreaRootLayer but
     get it directly from our view, we must not free it conversely to the Mac
     OS X version ! */
  
  if (transpImageHintLayer != NULL) {
    [transpImageHintLayer removeFromSuperlayer] ;
    transpImageHintLayer = NULL ;
  }
  
  holdPiece = NULL ;
  bordersHighlighted = FALSE ;
}



- (void) touchesMoved: (NSSet*) touches withEvent: (UIEvent*) event
{
  if (current_game != NULL) {
    UITouch *touch = [touches anyObject] ;
    CGPoint loc = [touch locationInView: self] ;
    if (holdPiece != NULL) {
      int x = (int) loc.x ;
      int y = (int) loc.y ;
      /* Ensure that it is impossible to move the hold piece beyond the limits
         of the view. This way, we are always sure that the user will be able to
         pick it later. */
      if ((x >= 0) && (x < maxViewWidth) && (y >= 0) && (y < maxViewHeight)) {
        movePieceTo (x, y, current_game, holdPiece) ;
        [self partialUpdateDisplay: current_game fromPiece: holdPiece] ;
      }
    }     /* End of if (holdPiece != NULL). */
  }
}



- (void) touchesBegan: (NSSet*) touches withEvent: (UIEvent*) event
{
  if (current_game != NULL) {
    UITouch *touch = [touches anyObject] ;
    CGPoint loc = [touch locationInView: self] ;
    holdPiece = findClosestPiece (current_game, (int) loc.x, (int) loc.y) ;
    if (holdPiece == NULL) {
      /* No piece is selected, so consider the user wants to scroll. */
      if (self.scrollEnabled == NO) self.scrollEnabled = YES ;
    }    /* End of if (holdPiece == NULL). */
    else self.scrollEnabled = NO ;
  }      /* End of if (current_game != NULL). */
}



- (void) touchesEnded: (NSSet*) touches withEvent: (UIEvent*) event
{
  if ((current_game != NULL) && (holdPiece != NULL)) {
    checkCoalescence (current_game, holdPiece) ;
    [self partialUpdateDisplay: current_game fromPiece: holdPiece] ;
    holdPiece = NULL ;
    if (isGameJustWon (current_game)) {
      /* Show a little congratulation message. */
      UIAlertView *alert =
        [[UIAlertView alloc]
         initWithTitle: @"Game ended" message: @"You won the game."
         delegate: nil cancelButtonTitle: @"Ok, I'm good"
         otherButtonTitles: nil] ;
      [alert show] ;  
      [alert release] ;
    }
  }
}



/** Makes all the GUI administation to prepare to launch a game already
    allocated (i.e. created or loaded). This function deals with the size of the
    window, of the views, the menus enabling, the root and hint layers
    allocation and calls the function that initialize layers for pieces.  */
- (void) setupViewForGame: (int) window_width
         withWindowHeight: (int) window_height
{
  /* Recover the root layer we insert in our custom view. But, conversely to
     the Mac OS X version, on iOS we do not allocate this root layer: we
     directly peek it from our view. So at cleanup-time we must not free it ! */
  playAreaRootLayer = self.layer ;
  
  /* Don't flip the layer's coordinates system. On iOS, coordinate system is
     already with origin at the top-left corner. */
  
  /* Set the size of the root view (i.e. the PlayAreaView, that is, ourself).
     The game creation routine will allocate a layout of this size. */
  [self setContentSize: CGSizeMake(window_width, window_height)] ;
  maxViewWidth = window_width ;
  maxViewHeight = window_height ;
    
  /* Create the semi-transparent layer in which we will display, if requested,
     the ghost of the picture to rebuild. We creat it here so it will be below
     all the layers created later and representing the pieces. */
  transpImageHintLayer = [CALayer layer] ;
  CGRect hintFrame ;
  /* Set the position and dimension of the layer. Center it in the window. */
  hintFrame.origin.x = (window_width - current_game->image_width) / 2 ;
  hintFrame.origin.y = (window_height - current_game->image_height) / 2 ;
  hintFrame.size.width = current_game->image_width ;
  hintFrame.size.height = current_game->image_height ;
  transpImageHintLayer.frame = hintFrame ;
  /* Don't flip the layer's coordinates system. On iOS, coordinate system is
     already with origin at the top-left corner. */
  transpImageHintLayer.opacity = 0.2 ;
  transpImageHintLayer.contents = nil ;
  [playAreaRootLayer addSublayer: transpImageHintLayer] ;
  
  [self initGamePiecesLayers: current_game] ;
}  



- (void) startNewGame: (UIImage*) image
      withHorizPieces: (int) nb_horiz_pieces
       withVertPieces: (int) nb_vert_pieces
      withScaleFactor: (float) scale_factor
      withWindowWidth: (int) window_width
     withWindowHeight: (int) window_height
{
  [self endCurrentGameAndCleanup] ;
  
  current_game =
    newGame
      (image.CGImage, window_width, window_height,
       nb_horiz_pieces, nb_vert_pieces, scale_factor) ;
  
  if (current_game != NULL) {
    bordersHighlighted = FALSE ;
    shuffleAllPieces (current_game) ;
    [self setupViewForGame: window_width withWindowHeight: window_height] ;
  }           /* End of if (current_game != NULL). */
  else
    displayMessage ("Error", "New game creation failed.", AC_Critical) ;  
}



- (IBAction) toggleImageHint
{
  [CATransaction flush] ;
  [CATransaction begin] ;
  [CATransaction setValue: (id) kCFBooleanTrue
                   forKey: kCATransactionDisableActions] ;  
  
  if (transpImageHintLayer.contents == nil)
    transpImageHintLayer.contents = (id) current_game->game_image ;
  else transpImageHintLayer.contents = nil ;
  [CATransaction commit] ;
}



- (IBAction) arrangePiecesOnBorders
{
  int i ;
  
  if (current_game != NULL) {
    arrangePiecesOnBorders (current_game) ;
    /* Refresh the display. */
    for (i = 0; i < current_game->nb_pieces; i++)
      [self partialUpdateDisplay: current_game
                       fromPiece: &current_game->pieces[i]] ;
  }
}


/** Toggles transparency (0.25 <-> 1.0) of pieces except borders. */
- (IBAction) highlightBorders
{
  int x, y ;
  int nb_cols, nb_lines ;
  float opacity ;
  
  if (current_game == NULL) return ;
  
  bordersHighlighted = !bordersHighlighted ;
  opacity = (bordersHighlighted ? 0.25 : 1.0) ;
  
  nb_cols = current_game->image_width / current_game->piece_width ;
  nb_lines = current_game->image_height / current_game->piece_height ;
  
  for (y = 1; y < nb_lines - 1; y++) {
    for (x = 1; x < nb_cols - 1; x++) {
      piecesLayers[(y * nb_cols) + x].opacity = opacity ;
    }
  }
}

@end
