/*
  SetupView.m

  Created by François Pessaux on 08/10/10.
  Copyright 2010. All rights reserved.
*/

#import "JigPlayViewController.h"
#import "SetupView.h"

#include "misc.h"
#include "alerts_n_errors.h"

@implementation SetupView




- (void) initGUIWithDefaultValues
{
  selectedImage = nil ;
  original_image_width = 0 ;
  original_image_height = 0 ;
  scaled_image_width = 0 ;
  scaled_image_height = 0;
  scale_factor = 1.0 ;
  nb_found_resolutions = 0 ;
  resolutions = NULL ;
  resolutionsTexts = nil ;
  
  widthLabel.text = @"-" ;
  heightLabel.text = @"-" ;
  averagePieceSizeLabel.text = @"-" ;
  resolutionsPicker.autoresizingMask = UIViewAutoresizingFlexibleWidth ;
  /* Note that default is NO. */
  resolutionsPicker.showsSelectionIndicator = YES ;
  /* This view is the data source and delegate. Don't forget to make the links
     under IBuilder. */
  resolutionsPicker.delegate = self ;
  resolutionsPicker.dataSource = self ;
  
  /* Add this picker to our view controller and make it visible. */
  resolutionsPicker.hidden = NO ;
  [self addSubview: resolutionsPicker] ;
  
  /* Disable the Ok button until the settings are correctly initialized. */
  [okButton setEnabled: NO] ;
}



#pragma mark -
#pragma mark UIPickerViewDataSource
- (NSInteger) numberOfComponentsInPickerView: (UIPickerView*) pickerView
{
  return (1) ; /* This makes our picker having 1 slot. */
}



- (NSInteger) pickerView: (UIPickerView*) pickerView
 numberOfRowsInComponent: (NSInteger) component
{
  return [resolutionsTexts count] ;
}



/* This method is responsible in redering the picker's cells in the view. */
- (NSString*) pickerView: (UIPickerView*) pickerView
             titleForRow: (NSInteger) row
            forComponent: (NSInteger) component
{
  NSString *returnStr = @"" ;
  if (pickerView == resolutionsPicker) {
    returnStr = [resolutionsTexts objectAtIndex: row] ;
  }
  return (returnStr) ;
}



#pragma mark -
#pragma mark UIPickerViewDelegate
- (void) pickerView: (UIPickerView*) pickerView
       didSelectRow: (NSInteger) row
        inComponent: (NSInteger) component
{
  if (pickerView == resolutionsPicker) {
    int index = [pickerView selectedRowInComponent: 0] ;
    if ((index < 0) || (index > nb_found_resolutions)) return ;
    averagePieceSizeLabel.text =
      [NSString stringWithFormat: @"%d",
       resolutions[index].average_piece_size] ;
  }
}



#pragma mark -
#pragma mark Resolutions Computation and Update

/* Returns TRUE if one resolution is available and selected by default. If none,
 then this return value must trigger the disactivation of the OK button. */
- (bool_t) updateResolutions
{
  int width, height ;
  int i ;
  
  [resolutionsTexts removeAllObjects] ;
  
  if (resolutions != NULL) {
    free (resolutions) ;
    resolutions = NULL ;
  }
  if (resolutionsTexts != nil) {
    [resolutionsTexts removeAllObjects] ;
    [resolutionsTexts release] ;
    resolutionsTexts = nil ;
  }
  
  if (scaleSwitch.on) {
    width = scaled_image_width ;
    height = scaled_image_height ;
  }
  else {
    width = original_image_width ;
    height = original_image_height ;
  }
  
  resolutions =
    computeRecommendedResolutions (width, height, &nb_found_resolutions) ;
  resolutionsTexts =
    [[NSMutableArray alloc] initWithCapacity: nb_found_resolutions] ;
  if ((resolutions != NULL) && (resolutionsTexts != nil)) {
    for (i = 0; i < nb_found_resolutions; i++) {
      struct resolution *r = &resolutions[i] ;
      NSString *s =
        [NSString stringWithFormat: @"%d x %d",
         r->nb_horizontally, r->nb_vertically] ;
      [resolutionsTexts insertObject: s atIndex: i] ;
    }     /* End of for (i = 0; i < nb_found_resolutions; i++). */
    
    
  }
  
  [resolutionsPicker reloadAllComponents] ;
  int index = [resolutionsPicker selectedRowInComponent: 0] ;
  if (index > -1) {
    /* Update the textfield telling the average piece size. */
    averagePieceSizeLabel.text =
      [NSString
        stringWithFormat: @"%d", resolutions[index].average_piece_size] ;
    return (TRUE) ;
  }
  else {
    averagePieceSizeLabel.text = @"-" ;
    return (FALSE) ;
  }
}



#pragma mark -
#pragma mark GUI callbacks

- (IBAction) scaleFactorSliderMoved:(id) sender
{
  scale_factor = scaleFactorSlider.value ;
  
  /* Update class internal variables. */
  scaled_image_width =
    (int) ((float) original_image_width * scale_factor) ;
  scaled_image_height =
    (int) ((float) original_image_height * scale_factor) ;
  /* No need to check if scaling is enabled since when it is not, the slider is
     not activable, hence it can't move, hence this routine can't be called. So
     if we are called, then scaling is allowed and the width and height we must
     display are really the scaled ones. */
  widthLabel.text = [NSString stringWithFormat: @"%d", scaled_image_width] ;
  heightLabel.text = [NSString stringWithFormat: @"%d", scaled_image_height] ;
  /* Update the suggested resolutions. */
  [okButton setEnabled: [self updateResolutions]] ;
}



- (IBAction) scaleSwitchChanged: (id) sender
{
  [scaleFactorSlider setEnabled: scaleSwitch.on] ;
  
  /* Do not set any value in the text fields showing the piece size if there
     is no selected image, i.e. if the image width or height is null. */
  if ((original_image_width <= 0) || (original_image_height <= 0)) return ;
  
  [okButton setEnabled: [self updateResolutions]] ;
}



- (void) imageSelectedCallback: (UIImage*) image
{
  /* Record the new image. */

  /* TODO: buggy, but there may be a memory leak.
  if (selectedImage != nil) [selectedImage release] ; */
  
  [image retain] ;
  selectedImage = image ;
  
  /* Extract and update the dimensions of the image. */
  CGSize imgSize = selectedImage.size ;
  original_image_width = imgSize.width ;
  original_image_height = imgSize.height ;

  /* Update class internal variables. */
  scaled_image_width =
    (int) ((float) original_image_width * scale_factor) ;
  scaled_image_height =
    (int) ((float) original_image_height * scale_factor) ;
  /* Allows scaling. */
  [scaleSwitch setEnabled: YES] ;
  
  /* Update text fields reporting the image size. */
  if (scaleSwitch.on) {
    [scaleFactorSlider setEnabled: YES] ;
    widthLabel.text = [NSString stringWithFormat: @"%d", scaled_image_width] ;
    heightLabel.text = [NSString stringWithFormat: @"%d", scaled_image_height] ;
  }
  else {
    [scaleFactorSlider setEnabled: NO] ;
    widthLabel.text =
      [NSString stringWithFormat: @"%d", original_image_width] ;
    heightLabel.text =
      [NSString stringWithFormat: @"%d", original_image_height] ;
  }

  /* Update the suggested resolutions and finally activate the OK button if at
     least one resolution was found and selected by default. */
  [okButton setEnabled: [self updateResolutions]] ;
}



- (IBAction) cancelButtonPressed: (id) sender
{
  /* Cleanup. */
  if (resolutionsTexts != nil) {
    [resolutionsTexts removeAllObjects] ;
    [resolutionsTexts release] ;
    resolutionsTexts = nil ;
  }  
  if (resolutions != NULL) {
    free (resolutions) ;
    resolutions = NULL ;
  }
  
  /* Go back to the intro view. */
  [viewController showIntroView] ;
}



- (IBAction) okButtonPressed: (id) sender
{
  int width, height, piece_width, piece_height, i ;
  
  i = [resolutionsPicker selectedRowInComponent: 0] ;
  if ((i < 0) || (i >= nb_found_resolutions)) {
    displayMessage
      ("Error. Please report.", "Unable to recover dimensions.", AC_Critical) ;
    return ;
  }
  int nb_horiz_pieces = resolutions[i].nb_horizontally ;
  int nb_vert_pieces = resolutions[i].nb_vertically ;

  if (scaleSwitch.on) {
    width = scaled_image_width ;
    height = scaled_image_height ;
  }
  else {
    width = original_image_width ;
    height = original_image_height ;
    /* To tell the game creation routine not to scale. */
    scale_factor = -1 ;
  }
  
  /* Compute the possibly cropped dimension due to number of pieces not
     being an exact divisor of the image dimension. */
  piece_width = width / nb_horiz_pieces ;
  piece_height = height / nb_vert_pieces ;
  width = nb_horiz_pieces * piece_width ;
  height = nb_vert_pieces * piece_height ;
  
  /* Extend the width and height so that there is room to move the pieces
     around. Adding 75% should be sufficient. */
  width = width + ((width * 3) / 4) ;
  height = height + ((height * 3) / 4) ;

  /* Cleanup. */
  if (resolutionsTexts != nil) {
    [resolutionsTexts removeAllObjects] ;
    [resolutionsTexts release] ;
    resolutionsTexts = nil ;
  }
  if (resolutions != NULL) {
    free (resolutions) ;
    resolutions = NULL ;
  }
  
  /* Finally, call the PlayAreaView method to start a new game. */
  [viewController startNewGameAndShowPlayAreaView: selectedImage
                                  withHorizPieces: nb_horiz_pieces
                                   withVertPieces: nb_vert_pieces
                                  withScaleFactor: scale_factor
                                  withWindowWidth: width
                                 withWindowHeight: height] ;
}

@end
